//---------------------------------------------------------------------------
//
//  Time-stamp: <2003-11-24 19:39:25 sutner> 
//
// :File:      atomscan.cc
// :Purpose:   simple scanner for atoms
// 
//---------------------------------------------------------------------------

#include  <iostream>
#include  <fstream>
#include  <sstream>

#include  "basic.h"
#include  "urlist.h"
#include  "useratom.h"

//--------------------------------------------------------------------------------

const int BLEN = 256;

static  char  symbuff[BLEN];
static  char  buffer[BLEN];
static  char  *bp;

static  bool  fill_buffer( istream& strmp )
{
    int c, i=0;
    bp = buffer;
    while( (c=strmp->get()) && c!='\n' && c!=EOF &&  ++i < BLEN )
	*bp++ = c;
    *bp = '\0';

    if( strmp->eof() || strmp->fail() )
	return 0;

    bp = buffer;
    skip_spaces();
    return 1;
}

void   skip_spaces()
{
    while( (*bp == ' ') || (*bp == '\t') ) bp++;
    if( *bp=='#')	// skip comment
	while( *bp ) bp++;	
}

void  skip_nl()
{
    skip_spaces();
    while (*bp == '\0') 	// end of line
    {
	secondary_prompt();
	fill_buffer();
    }
}

void   read_string( istream& strmp )
{
    char *q = symbuff;
    while( isalnum(*bp) )		// a string
	*q++ = *bp++;
    *q = '\0';
}


UList  read_atom_raw( istream& strmp )
{
    while( fill_buffer(strmp) )
    {
	if( *bp == '#' )		// comment, skip
	    continue;
	
	if( isalnum(*bp) )		// get header and dispatch
	    read_string();
    
	if( strcmp(symbuff,"SA") )   return read_sa_atom(strcmp);
	if( strcmp(symbuff,"FA") )   return read_fa_atom(strcmp);
	if( strcmp(symbuff,"DFA") )  return read_dfa_atom(strcmp);
	if( strcmp(symbuff,"CA") )   return read_ca_atom(strcmp);
	
    }
    
    return nil;			// should never get here
}


UList  read_sa_atom( istream& strmp )
{
    fill_buffer(strmp);
    
    char  *q = symbuff;		// write into symbuff

    if( isdigit(*bp) || *bp == '-' )		// an integer
    {
	*q++ = *bp++;		// copy minus or first digit
	while( isdigit(*bp) )
	    *q++ = *bp++;
	*q = '\0';
	return   INT;
    }
    
    if( *bp == '\"' )		// a quoted string
    {
	bp++;
	while( *bp != '\"' )
	    *q++ = *bp++;
	*q = '\0';
	bp++;
	return   STRT;
    }

    while( isalnum(*bp) )	// a symbol
	*q++ = *bp++;
    *q = '\0';
    
    TokenHTab::iterator  it;	// check for reserved word
    it = toktab.find(symbuff);
    if( it==toktab.end() )
	return  VAR;
    else
	return  (*it).second;
}

void  Parser::print_table()
{
    TokenHTab::iterator  it;
    for( it=toktab.begin(); it!=toktab.end(); ++it )
	cerr << (*it).first << "  " << (*it).second << endl;
    cerr << endl;
}


//--------------------------------------------------------------------------------
// fill buffer, scan and parse one block of input in interactive mode
// check for print flag

UrList  Parser::parse()
{
    PARSEMSSG("parse");

    dopr = 1;			// default: printing

    do				// wait for user input
    {
	primary_prompt();
	fill_buffer();
    } 
    while( !*bp && goflag );

    return read_input(1);	// parse input so far
}

UrList  Parser::parse_one()
{
    dopr = 0;			// printing off
    fill_buffer();
    return read_input(1);	// parse only one expression
}

//--------------------------------------------------------------------------------
// scan and parse one block of input in interactive mode
// distinguish
//      expressions
//      statements
// gobble trailing tokens

bool  Parser::is_stmt_token( Token t )
{
    return  (t==IF || t==WHILE || t==LBR || t==DEFUN || t==STOP );
}


UrList  Parser::read_input( bool nt )
{
    PARSEMSSG("read_input");

    if( nt ) get_token();
    
    UrList  val;
    
    if( is_stmt_token( token ) )
	val = read_stat();
    else
	val = read_expr();

    check_semicolon();    

    if( *bp )			// gobble trailing chars
    {
	error( "unexpected characters at end of line: ", bp );
    }
    return val;
}

//--------------------------------------------------------------------------------
// read statement
// use semicolon as statement 

UrList  Parser::read_stat( bool nt )
{
    PARSEMSSG("read_stat");

    if( nt ) get_token();

    switch( token )
    {
    case IF:        return read_if();
    case WHILE:     return read_while();
    case DEFUN:     return read_defun(1);
    case LBR:       return new CompAtom ( read_compstat(1) );

    case STOP:      goflag = 0; return nil;
	
    default:        return  read_exprstmt(0);
    }
}

//--------------------------------------------------------------------------------
// expressions:  atomic, lists, assignments

UrList  Parser::read_expr( bool nt )
{
    PARSEMSSG("read_expr");

    if( nt ) get_token();

    switch( token )
    {
    case LISTBEG:   return read_list(1); 
    case INT:       return read_int();
    case STRT:      return read_string();
    case ATOM:                                   // return read_atom();
    case OPER:                                   // return read_oper();
    case VAR:       return read_symbol();

    case LAST:      return  last_val;

    default:        error("bad expression token"); return  nil;
    }
}

//--------------------------------------------------------------------------------

UrList  Parser::read_exprstmt( bool nt )
{
    PARSEMSSG("read_exprstmt");

    UrList val = read_expr(nt);

    get_token();
    
    if( token != SCL )
    {
	error( "missing semicolon", bp );
	return nil;
    }
    return val;
}

//--------------------------------------------------------------------------------
// read blank separated list expression
// pre:    LPR already read
// post:   matching right parens all scanned 

UrList  Parser::read_list( bool nt )
{
    PARSEMSSG("read_list");

    if( nt ) get_token();
    
    if( token == LISTEND )	// list is empty
	return nil;

    UrList  val = read_expr();	// else: read first atom

    get_token();

    if( token==COMMAT )
    {
	skip_nl();		// allow entries on new line, needed ????
	return  val -= read_list(1);
    }
    
    if( token==LISTEND ) 
	return  val -= nil;
}

//--------------------------------------------------------------------------------
// read compound statement: semicolon separated list of statements
// pre:   left brace already scanned
//        nt==1:  read next, possibly right brace
// post:  will read matching right brace

UrList  Parser::read_compstat( bool nt )
{
    PARSEMSSG("read_compstat");

    skip_nl();
    if( nt ) get_token();
    
    if( token==RBR )	// empty compound statement
	return nil;

    UrList stmt  = read_stat();
    skip_nl();
    return  stmt -= read_compstat(1);
}

//--------------------------------------------------------------------------------
// pre: 
//        nt==0:  left paren already is token
//        nt==1:  try to read left paren
// post:  will read matching right paren

UrList  Parser::read_cond( bool nt )
{
    PARSEMSSG("read_cond");

    UrList cond;

    if( nt ) get_token(); 

    if( token != LPR )
    {
	error( "condition not well-formed 1", bp );
	return nil;
    }
    
    cond = read_expr(1);

    get_token(); 

    if( token != RPR )
    {
	error( "condition not well-formed 2", bp );
	return nil;
    }
    
    return cond;
}

//--------------------------------------------------------------------------------
// read if-then or if-then-else
// pre:  token IF already scanned

UrList  Parser::read_if() 
{
    PARSEMSSG("read_if");

    UrList cond, ifpart, elsepart;

    cond = read_cond(1);

    skip_nl();
    ifpart = read_stat(1);	// has gobbled ; or }

    skip_nl();
    get_token();
    
    if( token == FI )
	return  new IfAtom( cond, ifpart, nil );

    
    if( token != ELSE )
    {
	error("need `fi' or `else' in conditional");
	return  nil;
    }
    
    skip_nl();
    elsepart = read_stat(1);

    skip_nl();
    get_token();
    
    if( token == FI )
	return  new IfAtom( cond, ifpart, elsepart );

    error("need `fi' in conditional");
    return  nil;
}

//--------------------------------------------------------------------------------
// pre:   while token is read, but not conditional

UrList  Parser::read_while()
{
    PARSEMSSG("read_while");

    UrList cond, body;

    cond = read_cond(1);

    skip_nl();
    get_token();
    
    if( token == LBR )
	body = new CompAtom( read_compstat(1) );
    else
	body = read_stat();
    
    return  new WhileAtom( cond, body );
}

//--------------------------------------------------------------------------------
// pre:  defun token already scanned

UrList  Parser::read_defun( bool nt )
{
    PARSEMSSG("read_defun");

    UrList fname, argl, body;

    if( nt ) get_token();	// can nt==0 happen???
    
    if( token != VAR )
    {
        error( "defun:  missing function name", bp );
	return nil;
    }
    
    fname = read_string();

    get_token();
    if( token == LPR )
	argl = read_paralist(1);      // gobbles RPR
    else
    {
	error("no parameter list");
	return nil;
    }

    skip_nl();    
    body = read_stat(1);
    return new DefAtom( cvalue(fname), argl, body );
}

//--------------------------------------------------------------------------------

UrList  Parser::read_paralist( bool nt )
{
    PARSEMSSG("read_paralist");

    if( nt ) get_token(); 

    if( token == RPR )	// list is empty
	return nil;

    if( token != VAR )
    {
	error( "paralist: found bad parameter", bp );
	return nil;
    }
    
    UrList arg  = read_string();

    get_token();
    if( token == COMMAT )
	return  arg -= read_paralist(1);
    
    if( token == RPR )
	return  arg -= nil;
    
    error( "bad parameter list", bp );
    return nil;
}

//--------------------------------------------------------------------------------

UrList  Parser::read_exprlist( bool nt )
{
    PARSEMSSG("read_exprlist");

    if( nt ) get_token(); 

    if( token == RPR )	// list is empty
	return nil;

    UrList arg  = read_expr();

    get_token();
    if( token == COMMAT )
	return  arg -= read_exprlist(1);
    
    if( token == RPR )
	return  arg -= nil;
    
    error( "bad expression list", bp );
    return nil;
}

//--------------------------------------------------------------------------------
// can't get_token after var: if not ASSG, have to push it back

//#define  PARSEDEBUG

UrList  Parser::read_symbol()
{	
    PARSEMSSG("read_symbol");

#ifdef  PARSEDEBUG
	cerr << "read_sym:  " << symbuff << " :  ";
#endif

    UrList expr, var = new StringAtom(symbuff);

    skip_spaces();

    if( *bp == '=' )
    {	
	PARSEMSSG("read assg");
	bp++;

	expr = read_expr(1);
#ifdef  PARSEDEBUG
	cerr << "x = y  " << expr << endl;
#endif
	return   new AssgnExprAtom( cvalue(var), expr );
    }

    if( *bp == '(' )
    {	
	PARSEMSSG("read function");
	bp++;

				// special case: UrList of int's
	if( strcmp( cvalue(var), "IL" )==0 )
	{
	    expr = read_ilist();
#ifdef  PARSEDEBUG
	cerr << "ilist  " << expr << endl;
#endif
	    return expr;
	}
	
	expr = read_exprlist(1); // gobbles RPR

#ifdef  PARSEDEBUG
	cerr << "x(y)  " << length_ul(expr) << " : " << expr << endl;
#endif
	return   new FuncExprAtom( cvalue(var), expr );
    }

    extern SymTable  globOpTab;
    UrList ff = globOpTab.lookup( cvalue(var) );

#ifdef  PARSEDEBUG
    cerr << "glop  " << ff << endl;
    cerr <<  (( ff.isNil() )? "nil" : "not nil") << endl; 
#endif

    if( !ff.isNil() ) 
	return ff;
    else
	return new VarExprAtom( cvalue(var) );
}

//--------------------------------------------------------------------------------
// atomic components
//--------------------------------------------------------------------------------
// test for small ints very crude
// does not permit negative gmp numbers

UrList  Parser::read_int()
{
    PARSEMSSG("read_int");

    if( strlen(symbuff) >= 10  )
	return  new IntegerAtom( symbuff );

    if( symbuff[0] == '-' )
	return  new IntAtom( -atoi( symbuff+1 ) );
    
    return  new IntAtom( atoi(symbuff) );
}

//--------------------------------------------------------------------------------
// quoted strings

UrList  Parser::read_string()
{
    PARSEMSSG("read_string");
    return  new StringAtom( symbuff );
}


//--------------------------------------------------------------------------------
// CURRENTLY NOT USED
//--------------------------------------------------------------------------------

// //--------------------------------------------------------------------------------
// // pre:   operator symbol already scanned

// UrList  Parser::read_oper()
// {
//     PARSEMSSG("read_oper");

//     Token  tk = token;
//     UrList op, argl;

//     op = read_string();	

//     get_token();
//     if( token == LPR )
//     {
// 	argl = read_exprlist(1);      // gobbles RPR
// 	if( tk==ATOM )
// 	    return  new FuncExprAtom( cvalue(op), argl -= nil );
// 	else
// 	    return  new FuncExprAtom( cvalue(op), argl );
//     }

// 				// function atom, this works only for built-ins
// //    bp--;			// yuck, unget_token
//     extern SymTable  globOpTab;
//     UrList ff = globOpTab.lookup( cvalue(op) );

//     cerr << "read_oper:  " << cvalue(op) << "  " << ff << endl;

//     return ff;
// }

// //--------------------------------------------------------------------------------
// UrList  Parser::read_atom( bool nt )
// {
//     PARSEMSSG("read_atom");

//     if( nt ) get_token(); 

//     UrList  argl, atn = new StringAtom(symbuff);

//     get_token();
//     if( token == LBK )
//     {
// 	argl = read_atparlist(1);      // gobbles RBR
// 	return  create_atom( atn, argl );
//     }

//     error( "bad atom", bp );
//     return nil;
// }

// UrList  Parser::read_atparlist( bool nt )
// {
//     PARSEMSSG("read_atparlist");

//     if( nt ) get_token(); 

//     if( token == RBK )	// list is empty
// 	return nil;

//     UrList arg  = read_expr();	    // <<<< specialize to fast read 

//     get_token();
//     if( token == COMMAT )
// 	return  arg -= read_atparlist(1);
    
//     if( token == RBK )
// 	return  arg -= nil;
    
//     error( "bad atom para. list", bp );
//     return nil;
// }


//--------------------------------------------------------------------------------
// process an UrList<int>
// performs no error checking, make sure the input is properly formed
// does this save any time?
// static  stack<UrList,list<UrList> >  ilistStack;

UrList   Parser::read_ilist()
{
    int c, level, nn;
    stack<UrList,list<UrList> >  ilistStack;
    UrList tmp;
    ilistStack.push( nil );

    level = 0;
    do
    {
	skip_spaces();
	get_token();
	
        switch( token )
        {
//        case ' ':
//        case '\n':
        case COMMAT:
            break;
                
        case  LPR:
            ilistStack.push( nil );
            ++level;
            break;      

        case  RPR:
            tmp = reverse( ilistStack.top() );
            ilistStack.pop();
            ilistStack.top() = tmp -= ilistStack.top();
            --level;
            break;
      
        default:                // token INT, must be a plain unsigned integer
	    
            ilistStack.top() = new IntAtom( atoi(symbuff) ) -= ilistStack.top();
            break;      
        }
    } while ( level>0 );
    
    skip_spaces();
    get_token();
    assert( token==RPR );
    return  ilistStack.top().Head();
}
    
//--------------------------------------------------------------------------------

