//---------------------------------------------------------------------------
//
//  Time-stamp: <2003-11-24 18:09:08 sutner> 
//
// :File:      autom.cc
// :Purpose:   cmdline interface to some FSM operations
// 
//---------------------------------------------------------------------------


#include  "urlist.h"
#include  "listutil.h"
#include  "useratom.h"
#include  "util.h"
#include  "Clock.h"
#include  "fsm.h"
#include  "ioatom.h"
#include  "sgroup.h"

#include  <string>
#include  <vector>
#include  <list>
#include  <getopt.h>
#include  <iostream>
#include  <fstream>
#include  <sstream>

//--------------------------------------------------------------------------------

enum  Action 
{ SKIP, POWB, POWSB, ACCB, MINB, SSGB, SSGDCB, NORM, 
  CLSSB, CASAB, WELB, KERB, PRODU, PRODI, QUOB, CPB, FISC,  FISCH, ISOB };


istream  *instrp;
ostream  *ostrp;

const  char* usemsg = 
"a:      compute accessible part\n"
"c:      classify cellular automaton\n"
"C:      convert cellular automaton to semiautomaton\n"
"d:      copy input atom to output file\n"
"f:      compute Fischer data\n"
"F:      compute synchronizing word for minimal Fischer automaton\n"
"i:      compute product DFA (intersection) of two DFAs\n"
"I:      check if two DFAs are isomorphic\n"
"k:      compute kernel automaton of a semiautomaton\n"
"m:      compute minimal automton\n"
"n:      turn normalization off\n"
"p:      compute power automaton\n"
"P:      compute size of power automaton\n"
"u:      compute (left) quotient of two DFAs\n"
"s:      generate syntactic semigroup\n"
"S:      generate syntactic semigroup and D-class decomposition\n"
"t:      display timing\n"
"u:      compute product DFA (union) of two DFAs\n"
"w:      convert cellular automaton to Welch automata";


// output formatting

char  *Format::lp = "(";
char  *Format::rp = ")";
char  *Format::cm = ", ";

//--------------------------------------------------------------------------------


void  process_cmdline( int argc, char **argv, Action& act, bool& norm )
{
    if( argc == 1 )
    {
	cerr << "Usage:  " << argv[0] << " -acCdiIfFkmnpPqsStuw [infile] [outfile]" << endl;
	cerr <<  usemsg << endl;
	exit(0);
    }
    
    instrp = NULL;
    ostrp = NULL;
    act = SKIP;
    norm = 1;			// do normalize
    int   cc;
    
    while( ( cc = getopt(argc,argv,"acCdiIfFkmnpPqsStuw")) != EOF )
	switch( cc )
	{
	case 'a': act = ACCB; break;
	case 'c': act = CLSSB; break;
	case 'C': act = CASAB; break;
	case 'd': act = CPB; break;
	case 'i': act = PRODI; break;
	case 'I': act = ISOB; break;
	case 'f': act = FISC; break;
	case 'F': act = FISCH; break;
	case 'k': act = KERB; break;
	case 'm': act = MINB; break;
	case 'n': norm = 0; break;
	case 'p': act = POWB; break;
	case 'P': act = POWSB; break;
	case 'q': act = QUOB; break;
	case 's': act = SSGB; break;
	case 'S': act = SSGDCB; break;
	case 't': UListInitializer::showtime = 1; break;
	case 'u': act = PRODU; break;
	case 'w': act = WELB; break;
	    
	default:
	    break;
	}
    

    if( optind < argc )
	instrp = new ifstream( argv[optind++]  );

    if( optind < argc )
	ostrp = new ofstream( argv[optind] );

    if( !instrp )
	instrp = new istream(cin.rdbuf());

    if( !ostrp )
	ostrp = new ostream(cout.rdbuf());
    
}

//--------------------------------------------------------------------------------

void  make_accessible( bool norm )
{
    UList   M = read_atom_raw( *instrp );

    FA  fa;
    DFA  dfa;

    switch( M.atom_type() )
    {
    case  FAA:
	favalue( M ).accessiblePart( fa, norm );
	write_atom_raw( *ostrp, fa, norm );
	break;
	
    case  DFAA:
	dfavalue( M ).accessiblePart( dfa, norm );
	write_atom_raw( *ostrp, dfa, norm );
	break;

    default:
	cerr << "error, must be FA or DFA" << endl;
	exit(1);
	break;
    }

}

void  make_power_size()
{
    UList   M = read_atom_raw( *instrp );
    DFA  dfa;

    switch( M.atom_type() )
    {
    case  SAA:
	*ostrp << savalue( M ).powerAutomatonSize() << endl;
	break;
	
    case  FAA:
	*ostrp << favalue( M ).powerAutomatonSize() << endl;
	break;

    default:
	cerr << "error, must be SA or FA" << endl;
	exit(1);
	break;
    }
}

void  make_power( bool norm )
{
    UList   M = read_atom_raw( *instrp );
    DFA  dfa;

    switch( M.atom_type() )
    {
    case  SAA:
	savalue( M ).powerAutomaton( dfa, norm );
	break;
	
    case  FAA:
	favalue( M ).powerAutomaton( dfa, norm );
	break;

    default:
	cerr << "error, must be SA or FA" << endl;
	exit(1);
	break;
    }

    write_atom_raw( *ostrp, dfa, norm );
}

void  make_kernel( bool norm )
{
    UList   M = read_atom_raw( *instrp );

    if( M.atom_type() != SAA )
    {
	cerr << "error, must be SA" << endl;
	exit(1);
    }
    
    SA  sa;
    UList  R = range_i( 0, savalue( M ).numStates() -1 ) -= nil;
    savalue( M ).kernelAutomaton( sa, R, norm );
    
    write_atom_raw( *ostrp, sa, norm );
}

void  make_minimal( bool norm )
{
    UList   M = read_atom_raw( *instrp );

    DFA  d0, d1;

    switch( M.atom_type() )
    {
    case  SAA:
	savalue( M ).powerAutomaton( d0, norm );
	d0.minimalAutomaton( d1, norm );
	break;
    case  FAA:
	favalue( M ).powerAutomaton( d0, norm );
	d0.minimalAutomaton( d1, norm );
	break;
    case  DFAA:
	dfavalue( M ).minimalAutomaton( d1, norm );
	break;

    default:
	break;
    }
    write_atom_raw( *ostrp, d1, norm );
}

void  make_semigroup( int nn=0 )
{
    UList   M = read_atom_raw( *instrp );

    if( M.atom_type() != DFAA )
    {
	cerr << "error, must be DFA" << endl;
	exit(1);
    }
    
    SGroup  sg;
    DFA   &dfa = dfavalue( M );
    size_t n = dfa.numStates();
    size_t k = dfa.numSymbols();
    UList A = intarray_to_ulist( dfa.transitions() );

    uvector  gg( n * k );
    ulist_to_intarray( A, gg );
    SGroupGen  gs( n, k, gg );
    switch( nn )
    {
    case 0:  gs.generate( sg, SGroupGen::GENW ); break;
    case 1:  gs.generate( sg, SGroupGen::GENE ); break;
    case 2: 		    
    default: gs.generate( sg, SGroupGen::GEND ); break;
    }

    *ostrp << sg << endl;
    if( nn > 0 ) 
    {
	sg.print_witnesses( *ostrp );
	sg.print_equations( *ostrp );
    }
    
    if( nn == 2 ) 
	sg.print_dclasses( *ostrp );
}

void  make_product_dfa( bool ui, bool norm )
{
    UList   M1 = read_atom_raw( *instrp );
    UList   M2 = read_atom_raw( *instrp );

    if( M1.atom_type() != DFAA || M2.atom_type() != DFAA )
    {
	cerr << "error, must be DFA" << endl;
	exit(1);
    }
    
    UList   pars =  new_int(ui) -= new_int(norm) -= nil;
    UList   M = productDFA( M1, M2, pars );

    write_atom_raw( *ostrp, dfavalue( M ), norm );
}

void  make_quotient_dfa( bool norm )
{
    UList   M1 = read_atom_raw( *instrp );
    UList   M2 = read_atom_raw( *instrp );

    if( M1.atom_type() != DFAA || M2.atom_type() != DFAA )
    {
	cerr << "error, must be DFA" << endl;
	exit(1);
    }
    
    UList   M = quotientDFA( M1,  M2, new_int(norm) );

    write_atom_raw( *ostrp, favalue( M ), norm );
}

void  test_isomorphic( bool norm )
{
    UList   M1 = read_atom_raw( *instrp );
    UList   M2 = read_atom_raw( *instrp );

    if( M1.atom_type() != DFAA || M2.atom_type() != DFAA )
    {
	cerr << "error, must be DFA" << endl;
	exit(1);
    }
    
    DFA m1, m2;
    dfavalue(M1).accessiblePart( m1 );
    dfavalue(M2).accessiblePart( m2 );
    bool iso = (m1 == m2);

    *ostrp << iso << endl; 
}

void  make_ca_to_sa()
{
    UList   M = read_atom_raw( *instrp );

    if( M.atom_type() != CAA )
    {
	cerr << "error, must be CA" << endl;
	exit(1);
    }
    
    SA   sa;
    cavalue( M ).toSA( sa );
    write_atom_raw( *ostrp, sa );
}

void  make_ca_to_welch( bool norm )
{
    UList   M = read_atom_raw( *instrp );

    if( M.atom_type() != CAA )
    {
	cerr << "error, must be CA" << endl;
	exit(1);
    }
    
    SA   sa, sl, sr;
    cavalue( M ).toSA( sa );
    sa.welchAutomaton( sr, norm );
    sa.reverseAutomaton();
    sa.welchAutomaton( sl, norm );

    write_atom_raw( *ostrp, sl, norm );
    *ostrp << endl;
    write_atom_raw( *ostrp, sr, norm );
}

void  make_analyze()
{
    UList   M = read_atom_raw( *instrp );

    if( M.atom_type() != CAA )
    {
	cerr << "error, must be CA" << endl;
	exit(1);
    }
    
    *ostrp << cavalue( M ).analyze() << endl;
}


void  copy_atom( bool norm )
{
    UList   A = read_atom_raw( *instrp );
    write_atom_raw( *ostrp, A, norm );
}

void  make_fischer( bool norm )
{
    UList   M = read_atom_raw( *instrp );

    if( M.atom_type() != SAA )
    {
	cerr << "error, must be SA" << endl;
	exit(1);
    }
    
    SA  sa;
    III_Triple  pmf = savalue( M ).toFischer( sa, norm );
    if( norm )
	*ostrp << pmf << endl;
    else
	write_atom_raw( *ostrp, sa, norm );
}


static inline int  comp_pairing( int n, int x, int y )
{
    int a = ( y < x ) ? y : x;
    int b = ( y > x ) ? y : x;
    return  n * a - ((a*(a+1)) >> 1) + b;
}

UList  make_synchronize_fischer( bool norm )
{
    UList  M = read_atom_raw( *instrp );
    
    if( M.atom_type() != SAA )
    {
	cerr << "error, must be SA" << endl;
	exit(1);
    }
    return  synchFischer( M );
}

    
    
//--------------------------------------------------------------------------------


int main( int argc, char **argv )
{
    bool    norm;
    Action  act;
    process_cmdline( argc, argv, act, norm );

    switch( act ) 
    {
    case ACCB:     make_accessible( norm ); break;
    		   
    case POWB:     make_power( norm ); break;
    		   
    case POWSB:    make_power_size(); break;
    		   
    case MINB:     make_minimal( norm ); break;
    		   
    case SSGB:     make_semigroup(); break;
    		   
    case SSGDCB:   make_semigroup( 2 ); break;
    		   
    case CASAB:    make_ca_to_sa(); break;
    		   
    case CLSSB:    make_analyze(); break;
    		   
    case WELB:     make_ca_to_welch( norm ); break;
    		   
    case KERB:     make_kernel( norm ); break;
    		   
    case PRODI:    make_product_dfa( 1, norm ); break;

    case PRODU:    make_product_dfa( 0, norm ); break;

    case QUOB:     make_quotient_dfa( norm ); break;

    case FISC:     make_fischer( norm ); break;

    case CPB:      copy_atom( norm ); break;

    case FISCH:    make_synchronize_fischer( norm ); break;

    case ISOB:     test_isomorphic( norm ); break;

    default:
	cerr << "no instruction, terminating ... " << endl;
    }

    delete instrp;
    delete ostrp;

    return EXIT_SUCCESS;
}

