//--------------------------------------------------------------------------------
//
// Time-stamp: <2003-11-24 17:37:14 sutner>
// 
// :File:     cacode.c
// :Purpose:  operations for one-dimensional cellular automata 
//            
//--------------------------------------------------------------------------------

#include  <cstdio>
#include  <cstdlib> 

#include  "basic.h"  
#include  "fsm.h"  
#include  "ca.h"  
#include  "util.h"

//------------------------------------------------------------------------------
//  CA procedures
//------------------------------------------------------------------------------

// pre:  A has size = capacity = k^w
void  CA::makeLabels( uvector& A ) const
{
    Fill( A, 0U );
    integer_to_array( rule(), A, numSymbols() );
}

// conversion CA to SA 
void  CA::toSA( SA& sa ) const 
{
    size_t w  = width();
    size_t k  = numSymbols();
    size_t nt = ipow(k,w);		// rule size == number of transitions
    size_t ns = nt/k;		        // number of states 
    size_t i  = 0; 
    sa.set( ns, k );
    sa.transitions().resize(0);
   
    uvector lab(nt);
    makeLabels( lab );

    if( k==2 )			// speed up binary case 
    {
	size_t mask=ns-1; 
	for( i=0; i<nt; i++ )
	    sa.transitionsInsert( (i>>1)&mask, lab[i], i&mask );
	sa.sortTransitions();
	return;
    }
   
    size_t y  = 0, ii; 
    for( size_t x=0; x<ns; x++ )
	for( ii=0; ii<k; ii++ )
	{
	    sa.transitionsInsert( x, lab[i++], y );
	    y = (y+1) % ns;
	}

    sa.sortTransitions();
    return;
}

//-------------------------------------------------------------------------
// compute orbit of a given pattern, see caevol.cc

void  evolution_ca( const CA&, const uvector&, const int, const int, uvector& );

void   CA::orbit( const uvector& pat, const int t, const int cyc, 
		  uvector& orb ) const
{
    evolution_ca( *this, pat, t, cyc, orb );
}

//-------------------------------------------------------------------------
// determine type of linear CA 

int  CA::analyze() const 
{				// discard if not even 1-balanced
   if( !balancedQ(1) )
     return 0;

   SA    sa;
   toSA( sa );			// convert to de Bruijn automaton
   return sa.analyzeSCC();
}

//------------------------------------------------------------------------------
//  de Bruijn graphs and cellular automata
//
//  construct de Bruijn graph B(w,k,lab): 
//  w: width of de Bruijn, NOT of CA, k: alphabet size, lab: labels;
//  lab must have size k^(w+1)  
//------------------------------------------------------------------------------

static  void bin_deBruijn_graph( const CA& ca, Graph& gr );

void   CA::toGraph( Graph& gr ) const 
{
    size_t w = width();
    size_t k  = numSymbols();
    size_t nn = ipow(k,w);	// rule size 
   
    if( k==2 )			// binary case 
    {
	bin_deBruijn_graph( *this, gr );
	return;
    }
   
    uvector  lab(nn);
    makeLabels( lab );

    gr.initialize( nn, k*nn, k );

    size_t ll=0,y=0;
    for( size_t x=0; x<nn; x++ )
	for( size_t i=0; i<k; i++ )
	{
	    gr.insertEdge( x, y, lab[ll++] );	
	    y = (y+1) % nn;
	}

    return;
}

//  special case: binary de Bruijn graph 

static void bin_deBruijn_graph( const CA& ca, Graph& gr )
{
    size_t w  = ca.width() - 1;
    size_t nn   = 1 << w;
    size_t mask = nn - 1;

    gr.initialize( nn, 2*nn, 2 );

    uvector lab(nn);
    ca.makeLabels( lab );

    size_t ll=0;
    for ( size_t i=0; i<2*nn; i++ )
	gr.insertEdge( (i>>1)&mask, i&mask, lab[ll++] );

    return;
}


//------------------------------------------------------------------------------
// balance testing for CAs
//------------------------------------------------------------------------------
// 1-balance for arbitrary/binary CAs
static int  one_balancedQ( const CA& ca );
static int  bin_r_balancedQ( const CA& ca, size_t r );

int  CA::balancedQ( int r ) const 
{
    if( r==1 )
	return one_balancedQ( *this );

    size_t k = numSymbols();
    if( k==2 )
	return  bin_r_balancedQ( *this, r );
   
    size_t w    = width();
    size_t nn   = ipow(k,w);
    size_t kr   = ipow(k,r);
    size_t kw1  = ipow(k,w-1);
    size_t bnd  = kr * kw1;
    size_t kw   = k * kw1;
    vector<size_t> count( kr );
    Fill( count, 0 );

    uvector lab(nn);
    makeLabels( lab );

    size_t i,x,xx,c,kp;
    for( xx=x=0; x<bnd; xx = ++x )
    {
	c  = 0;
	kp = 1;
	for( i=0; i<(size_t)r; i++ )
	{
	    c  += lab[xx % kw]*kp;
	    kp *= k;
	    xx /= k;
	}
	count[c]++;
	if( count[c]>kw1 ) return 0;
    }
   
    return 1;
}

static int  one_balancedQ( const CA& ca ) 
{
    size_t k  = ca.numSymbols();
    size_t w  = ca.width();
    size_t nn = ipow(k,w);
    size_t nn0 = nn/k;
    size_t i;
   
    uvector lab(nn);
    ca.makeLabels( lab );

    uvector  cnt(k);   
    Fill( cnt, 0U );
    
    for( i=0; i<nn; i++ )
	cnt[lab[i]]++;

    size_t bal=1;
    for( i=0; i<k && bal; i++ )
	bal = ( cnt[i]==nn0 );

    return bal;
}


int  bin_r_balancedQ( const CA& ca, size_t r ) 
{
    size_t w    = ca.width();
    size_t mask = (1<<w) - 1;
    size_t bnd  = 1<<(w+r-1);
    size_t kw   = 1<<(w-1);
    vector<size_t>  count( 1<<r );
    Fill( count, 0 );
    
    uvector lab(1<<w);
    ca.makeLabels( lab );

    size_t i,x,xx,c,kp;
    for( xx=x=0; x<bnd; xx = ++x )
    {
	c  = 0;
	kp = 0;
	for( i=0; i<r; i++ )
	{
	    c |= lab[xx & mask] << kp;          // ca.labels() 
	    kp++;
	    xx >>= 1;
	}
	count[c]++;
	if( count[c]>kw ) return 0;
    }   
    return 1;
}



//----------------------------------------------------------------------
//  base k counter,  n digits
//  MSD is d0, only needed for composition of CA

class Counter: public vector<size_t>
{
    size_t  _base;		// actually base - 1

public:
   
    Counter() {}
    Counter( size_t n, size_t k ): vector<size_t>(n), _base(k-1)
    {
	Fill( *this, 0U );
    }
    
    bool  increment()
    {
	int  i = size()-1;
	while( i>=0 && (*this)[i]==_base )
	    (*this)[i--] = 0;

	if( i>=0 )
	{
	    (*this)[i]++;
	    return 1;
	}
	return 0;
    }
};


void   CA::compose( const CA& c2, CA& cc ) const 
{
    size_t k  = numSymbols();
    size_t w1 = width();		// width c1 = *this
    size_t w2 = c2.width();	        // width c2
    size_t w  = w1 + w2 - 1;	        // width cc
    size_t nn = ipow( k, w  );
    size_t n1 = ipow( k, w1 );
    size_t n2 = ipow( k, w2 );

    uvector lab1( n1 );
    uvector lab2( n2 );
    uvector lab( nn );
    Fill( lab, 0U );
    
    makeLabels( lab1 );
    c2.makeLabels( lab2 );

    size_t i, j, b, bb; 

    Counter cnt(w,k);		// count from 0 to nn-1: local confs for cc
    j = 0;   
    do
    {
	b = 0;
	for( i=0; i<w1; i++ )	// first w1 symbols
	    b = k*b + cnt[i];
	bb = lab1[b];		//labels()[kw1-b-1];
      
	for( ; i<w; i++ )		// remaining symbols
	{
	    b  = (b*k) % n1 + cnt[i];
	    bb = bb*k + lab1[b];	//labels()[kw1-b-1];
	}
      
	lab[j++] = lab2[bb];	// c2.labels()[kw2-bb-1];
    } 
    while( cnt.increment() );

				// convert to Integer
    Integer r;
    array_to_integer( lab, r, k );
    cc.set( r, w, k );
}


ostream&  operator<<( ostream& out, const CA& ca )
{
   out << "CA[ " << ca.rule() << COMMA << ca.width() << COMMA; 
   out << ca.numSymbols() << "]";
   return out;
}

//----------------------------------------------------------------------

