//--------------------------------------------------------------------------------
//
// Time-stamp: <2003-11-24 17:36:18 sutner>
// 
//
// :File:     caevol.cc
// :Purpose:  evolution of CA, transient_periodic_ca only for binary CA's
//            NB: this version uses a trie, it is not memory-less
//            
//--------------------------------------------------------------------------------

#include  <iomanip>
#include  <cstdlib>
#include  "ca.h"
#include  "basic.h"

//----------------------------------------------------------------------
// defined herein 

void  evolution_ca( const CA& ca, const uvector& pat, int t, int cyc, uvector& orb );

void  transient_periodic_ca( int&  tr, int& pp, CA&  ca, uvector&  pat, 
			     int tmax, int cyc=0 );

static void   initialize_grid( int** grid, const uvector& pat, int w, int t, int cyc );
static void   initialize_configurations( int **pat0, int **pat1, int& stcnt, 
					 const uvector& pat, int w, int cyc );
static void   one_step( int *pat0, int *pat1, uvector& rule, int& stcnt, 
			int n, int w, int k, int kw, int cyc );
static void   run_ca( int *grid, uvector& rule, int n, int t, int w, int k, int kw, int cyc );


// static void   print_pat( int* pat, int n, int j );


//------------------------------------------------------------------------
// note: width of grid size is 2*radius + n 
// border region is initialized to 0 for fixed boundary conditions
// and to the w0 last values from the other end for cyclic conditions
//
// allocates grid
// 
static void   initialize_grid( int** grid, const uvector& pat, int w, int t, int cyc )
{
    int i, w0 = w/2, n = pat.size(), *ip, *op;
   
    *grid = (int*)calloc( (2*w0+n) * (t+1), sizeof(int) );
   
    op = *grid + w0;		// copy pat into first row of grid
    for( i = 0; i < n; i++)
	*op++ = pat[i]; 

    if( cyc )			// fix boundary conditions, if necessary 
    {
	ip = *grid + w0;
	op = ip + n;
	for( i = 0; i < w0; i++)
	    *op++ = *ip++; 

	ip = *grid + n;
	op = *grid;
	for( i = 0; i < w0; i++)
	    *op++ = *ip++; 
    }
}


//
// allocates pat0, pat1
//
static void   initialize_configurations( int **pat0, int **pat1, int& stcnt, 
					 const uvector& pat, int w, int cyc )
{
    int i, w0 = w/2, n = pat.size(), *ip, *op;
    stcnt = 0;
   
    *pat0 = (int*)calloc( 2*w0+n, sizeof(int) );
    *pat1 = (int*)calloc( 2*w0+n, sizeof(int) );
   
    op = *pat0 + w0;		// copy pat into pat0
    for( i = 0; i < n; i++)
	*op++ = pat[i]; 

    if( cyc )			// fix boundary conditions, if necessary 
    {
	ip = *pat0 + w0;
	op = ip + n;
	for( i = 0; i < w0; i++)
	    *op++ = *ip++; 

	ip = *pat0 + n;
	op = *pat0;
	for( i = 0; i < w0; i++)
	    *op++ = *ip++; 
    }
}


//
// toggle next configuration between pat0 and pat1
//
static void   one_step( int *pat0, int *pat1, uvector& rule, int& stcnt, 
			int n, int w, int k, int kw, int cyc )
{
    int i, j, s, w0, nn, *ip, *op, *isrc, *itrg;
   
    w0 = w/2;
    nn = n + 2*w0;
   
    isrc = (stcnt) ? pat1 : pat0;   // if(stctn) pat1 ==> pat0 else pat1 ==> pat0
    itrg = (stcnt) ? pat0 : pat1;
 
    ip = isrc;
    op = itrg + w0; 

    s = 0;			// set up sweep
    for( j = 0; j < w; j++ )
	s = k*s + *ip++;
    *op++ = rule[s];
   
    for( i=1; i < n; i++)		// execute sweep
    {
	s = (k*s) % kw + *ip++;
	*op++ = rule[s];
    }
      
    if( cyc )			// periodic boundary conditions
    {				// copy to right filler
	ip = itrg + w0;
	op = ip + n;
	for( i = 0; i < w0; i++)
	    *op++ = *ip++; 
				// copy to left filler
	ip = itrg + n;
	op = itrg;
	for( i = 0; i < w0; i++)
	    *op++ = *ip++; 
    }
    else			// fixed boundary conditions
    {
	op = itrg + n + w0;
	for( i = 0; i < w0; i++)
	    *op++ = 0; 
      
	op = itrg; 
	for( i = 0; i < w0; i++)
	    *op++ = 0; 
    }
}

//------------------------------------------------------------------------
// generate configurations  X_0,...,X_t   (including generation t)
//
// uses:   grid
// 

static void   run_ca( int *grid, uvector& rule, int n, int t, 
		      int w, int k, int kw, int cyc )
{
    int i, j, s, step, w0, nn, *ip, *op;
   
    w0 = w/2;
    nn = n + 2*w0;
   
    for( step=0; step<t; step++ )
    {
	ip = grid + step*nn;
	op = ip + nn + w0;
      
	s = 0;
	for( j=0; j<w; j++ )
	    s = k*s + *ip++;
	*op++ = rule[s];

	for( i=1; i < n; i++)
	{
	    s = (k*s) % kw + *ip++;
	    *op++ = rule[s];
	}
      
	if( cyc )			// periodic boundary conditions
	{
	    ip = grid + (step+1)*nn + w0;
	    op = ip + n;
	    for( i = 0; i < w0; i++)
		*op++ = *ip++; 
	 
	    op = grid + (step+1)*nn;
	    ip = op + n;
	    for( i = 0; i < w0; i++)
		*op++ = *ip++; 
	}
	else                 // fixed boundary conditions
	{
	    op = grid + (step+1)*nn + n + w0;
	    for( i = 0; i < w0; i++)
		*op++ = 0; 
	 
	    op = grid + (step+1)*nn;
	    for( i = 0; i < w0; i++)
		*op++ = 0; 
	}
//      print_pat( grid + nn*step, nn, step );
    }
}


//--------------------------------------------------------------------------
// compute evolution of a configuration on a 1-dim CA 
// ca:   cellular automaton
// pat:  starting configuration 
// t:    number of steps
// cyc:  ==1  periodic boundary conditions 
//       ==0  fixed boundary conditions 
//
// frees:  grid 
//       
void  evolution_ca( const CA&  ca, const uvector&  pat, int t, int cyc, uvector& orb )
{

    int    i, j, n, ww, w0, w, k, kw;
    int   *grid, *ip;
   
    w  = ca.width();
    w0 = w/2;
    ww = 2*w0;
    k  = ca.numSymbols();
    n  = pat.size();
    kw = ipow( k, w );

    uvector  rule(kw);
    ca.makeLabels( rule );

    initialize_grid( &grid, pat, w, t, cyc );

    run_ca( grid, rule, n, t, w, k, kw, cyc );  

    // careful with boundary strips
//    for( i = 0; i <= t; i++ )
//       print_pat( grid + nn*i + w0, n, i );

    orb.resize( (t+1)*n );	        // copy only central strip to output array
    uvector::iterator   itr = orb.begin();
    for( i = 0, ip = grid + w0; i <= t; i++, ip += ww )
	for( j = 0; j < n; j++ )
	    *itr++= *ip++;
   
    free(grid);
}

//--------------------------------------------------------------------------
// find transient/period in the evolution of a configuration on a 1-dim CA
// ONLY for alphabet k==2 !!!
// ca:   cellular automaton
// pat:  starting configuration 
// tmax: max number of steps
// cyc:  ==1  periodic boundary conditions 
//       ==0  fixed boundary void 
// returns period 0 for failure
//
// frees:  pat0, pat1
// 

void  transient_periodic_ca( int&  tr, int& pp, CA&  ca, 
			     uvector&  pat, int tmax, int cyc )
{
   int    n, nn, w0, w, k, kw, step; 
   int    *ip;
   int    *pat0, *pat1;
   int stcnt;

   w = ca.width();
   k = ca.numSymbols();
   assert( k==2 );		// bin tries cannot cope with k>2
   n = pat.size();
   kw = ipow( k, w );
   
   w0 = w/2;
   nn = n + 2*w0;

   uvector  rule(kw);
   ca.makeLabels( rule );
   
   bTrie   tree( n );		// tree of depth n
   bTrieP  pt;
   
   initialize_configurations( &pat0, &pat1, stcnt, pat, w, cyc );
   
   tr = pp = 0;
				// careful with boundary strip
   tree.probe_branch( pat0 + w0, pt ); 
// print_pat( pat0+w0, n, 0 );

   for( step = 1; step <= tmax; step++ )
   {
      one_step(pat0,pat1,rule,stcnt,n,w,k,kw,cyc);

      if( stcnt )
	ip = pat0 + w0;
      else
	ip = pat1 + w0;

// print_pat( ip, n, step );
      
      if( tree.probe_branch( ip, pt ) )
      {
	 tr = pt->content();
	 pp = step - tr;
	 return;
      }

      stcnt = !stcnt;
   }
   
   free(pat0);
   free(pat1);
}

//----------------------------------------------------------------------
// eof
//----------------------------------------------------------------------
