//---------------------------------------------------------------------------
//
//  Time-stamp: <99/01/05 16:39:37 sutner> 
//
// :File:      Random.h
// :Purpose:   A random number generator, implemented as a function object.
//             See Stroustrup, 3rd edition, 22.7  and
//             Miller & Park, CACM, 31 (1988) 10.
//
//             Seed either manual (s > 0) or taken from time.h for s = 0.
//
//             RandGen  g;
//             can be used for floats in [0,1] or integers 0 <= x < 2^31.
//
//             URand  g( n );
//             produces uniform random numbers in  0,1,...,n-1, hopefully.
//
//             Does not provide any other distributions.
//             
//---------------------------------------------------------------------------

#ifndef  _RANDOM_H_ 
#define  _RANDOM_H_


#include  <time.h>

//--------------------------------------------------------------------------------
// random number generator implemented as a function object
// generate() core function: produces the next random number
// frand() and irand() convert to floats and integers

class RandGen
{
    long  randx;		// state: last random number
    
public:

    RandGen( long s=0 ):  randx( s ? s : time(0) ) {}

    long  seed( long s ) { return randx = s; }
    long  show_seed()    { return randx; }
    
    double  frand();
    int     irand();
    int     operator()() { return irand(); }

private:
    int  generate();

};


class  URand: public  RandGen
{
    int  nn;
    
public:
    
    URand( int n=2, long s=0 ): RandGen(s), nn(n) {}

    int  operator()() 
    { 
	int r = int( nn * frand() ); // casting to an int
	return  (r==nn) ? nn-1 : r;
    }
};

//--------------------------------------------------------------------------------


int  RandGen::generate()
{ 
				    // some magic numbers
    const long	m = 2147483647;     // modulus = 2^31 - 1, a Mersenne prime
    const long	g = 16807;          // a generator of Z/m
    const long	q = 127773;         // m div g
    const long	r = 2836;           // m mod g
				    // the next number
    long t  = g * (randx % q) - r * (randx / q);
    return  randx = (0 < t)? t : t + m;   // fix sign 
}

double RandGen::frand()
{ 
    return ( generate() & 0x7FFFFFFF ) / 2147483647.0;
}


int  RandGen::irand()
{
    return  ( generate() & 0x7FFFFFFF );
}


#endif
