//--------------------------------------------------------------------------------
//
// Time-stamp: <2003-11-24 18:37:09 sutner>
// 
// File:     bitvect.h
// Purpose:  -- BitVector's using arrays of size_ts 
//           -- basic linear algebra over GF(2)
//           -- same for 2-dim matrices
//           -- cellular automata stuff for 90 and 150
//           
//--------------------------------------------------------------------------------

#ifndef _BITVECT_H
#define _BITVECT_H

#include  <stdlib.h>
#include  <iostream.h>
#include  <vector>

//----------------------------------------------------------------------

#define    ON_CHAR       '*'
#define    OFF_CHAR      '-'
#define    BLOCKSIZE     32	// assume 32-bit ints
#define    ONE_BLOCK     ~0U
#define    LOWBIT        1U
#define    HIBIT         (1U<<(BLOCKSIZE-1))

//----------------------------------------------------------------------

				// for initialization of vectors, f( i, numbits ) 
typedef int (*VEC_FN)(int, int);           
				// ditto for matrices
typedef int (*MAT_FN)(int, int, int, int);
 

//----------------------------------------------------------------------
// BitVector
// vectors over GF(2), represented by arrays of unsigned integers
// use 32-bit parallelism
// blocks are filled right-to-left, so block 0 is rightmost  

class  BitVector
{
private:

    int    numbits;		// number of actual bits in vector 
    int    numblocks;		// number of ints, construed as blocks 
    int    uselast;		// number of occupied bits in last block 
    size_t  *blocks;		// array of unsigned ints    
   
public:
   
    // Initialize/Finalize	
    BitVector();			// no allocation
    BitVector(int n);			// bits not initialized
    BitVector(int n, size_t iniblock);	// each block initialized

    ~BitVector();

    void destroy();			// delete array

    void initialize(int n );		// initialize to n bits
    void initialize(int n, int b );	// initialize to n bits, set to b
    void initialize(int n, VEC_FN initializefn); // same, but use initializefn        


    // Copy    
    BitVector( const BitVector& );
    BitVector operator=( const BitVector& );

    // Access for whole 32-bit blocks 
          size_t& operator[](int n)       { return blocks[n];}
    const size_t& operator[](int n) const { return blocks[n];} 


    size_t  get(int n) const;
    int   getbit(int i) const;	        // read ith bit 
    void  setbit(int i, int b=1);	// set ith bit to b 
   
    // Attributes
    int   size() const;			// number of bits 
    int   numBlocks() const;		// number of blocks 

    // Input/Output
    friend  ostream&  operator<<( ostream& out, const BitVector& bv );


    // GF(2) and CA stuff

    BitVector operator+( const BitVector& ) const;    // addition is xor 
    int operator*( const BitVector& ) const;          // multiplication is dot product 

    // cyclic or fixed boundary conditions
    BitVector  left_shift( int cyc=1 ) const;
    BitVector  right_shift( int cyc=1 ) const;
    BitVector  sigma( int cyc=1 ) const;	
    BitVector  sigma_plus( int cyc=1 ) const;	
};					   // end class BitVector

inline  int BitVector::getbit(int n) const 
{
   return  ( size_t(blocks[n/BLOCKSIZE]) & (1U << (n%BLOCKSIZE))) ? 1 : 0;
}

inline  void BitVector::setbit(int n, int b)
{
   if( b )
       blocks[n/BLOCKSIZE] |=  (1U << (n%BLOCKSIZE));
   else
       blocks[n/BLOCKSIZE] &= ~(1U << (n%BLOCKSIZE));
}


//---------------------------------------------------------------------------
// 0/1 matrices
// currently implemented as array of BitVectors, should probably be flattened
// total number of ints in matrix:  numrows x numblocks
//---------------------------------------------------------------------------

class  BitMatrix
{
private:
   
    vector<BitVector*>  _rows;	      // array of BitVectors
    size_t             _numcols;      // number of bits per row
   
public:
   
    BitMatrix() {}
    BitMatrix( int n );
    BitMatrix( int n, int m );
    BitMatrix( int n, int m, int bb );
    BitMatrix( int nrows, int ncols, size_t iniblock );

    ~BitMatrix();

    // Attributes
    int  numRows()    const { return _rows.size(); }
    int  numColumns() const { return _numcols; }
    void size(int& nrows, int& ncols);	   // based on number of bits 

    // Copy    
    BitMatrix( const BitMatrix& );
    BitMatrix& operator=( const BitMatrix& );

    // Output
    friend ostream& operator<<( ostream& out, BitMatrix& bm );

    // Access
          BitVector*  operator[]( int i )       { return  _rows[i]; }
    const BitVector*  operator[]( int i ) const { return  _rows[i]; }

    // initialize bits using inifn        needed ????
    void  initialize( int nrows, int ncols, MAT_FN inifn ); 

    // Operations 
    BitMatrix  operator+( BitMatrix& );	   // addition is xor 
    BitVector  operator*( BitVector& );	   // A.X  
    int        matrixRank();		   // compute rank of matrix 
    BitMatrix  matrix_inverse();		   // compute A^-1 
    BitMatrix  left_shift_fixed();
    BitMatrix  right_shift_fixed();
    BitMatrix  sigma_fixed();		
    BitMatrix  sigma_plus_fixed();

    // helpers
private:
    void swap_rows(int i, int j);
    void add_rows(int i, int j);

};



#endif // _BITVECT_H

