//---------------------------------------------------------------------------
//
//  Time-stamp: <2003-11-24 18:41:55 sutner> 
//
//  :File:      pairs.h
//  :Purpose:   Classes and operations for pairs and triples,
//             more luxurious than the STL verion. 
//             Includes non-standard sorting operations. 
//---------------------------------------------------------------------------


#ifndef _PAIRS_H
#define _PAIRS_H

#include  <iostream>
#include  "stl-wrap.h"

//---------------------------------------------------------------------------
//  pairs
//  pre:  types A, B must have value semantics
//        default constructor should be reasonable
//---------------------------------------------------------------------------

template<class A, class B> class Pair
{
    A  eins;
    B  zwei;
   
public:

    Pair(): eins(A()), zwei(B()) {}
    Pair( const Pair& src ): eins(src.eins), zwei(src.zwei) {}

// does not work in g++ 2.8.1
//    template<class A, class B>
//    Pair( const Pair<A,B>& src ): eins(src.eins), zwei(src.zwei) {}

    Pair( const A& aa, const B& bb ): eins(aa), zwei(bb) {}

    A&        first()        { return eins; } 
    B&        second()       { return zwei; }
    const A&  first()  const { return eins; } 
    const B&  second() const { return zwei; }
    void      get( A& aa, B& bb ) const       { aa = eins;bb = zwei; } 
    void      set( const A& aa, const B& bb ) { eins = aa;zwei = bb; } 

    Pair&  operator=( const Pair& src );
    bool   operator<( const Pair& src ) const;
    bool   operator==( const Pair& src ) const;
    
    friend ostream& operator<<<A,B>(ostream& out, const Pair& pp);
};


//--------------------------------------------------------------------------------
// free functions to create pairs

template<class A, class B>
Pair<A,B>  to_pair( const A& a, const B& b )
{
    return  Pair<A,B>(a,b);
}

//--------------------------------------------------------------------------------

template<class A, class B>
inline  Pair<A,B>&  Pair<A,B>::operator=( const Pair<A,B>& src )
{
    if( &src==this ) return *this;

    eins = src.eins;
    zwei = src.zwei;

    return *this;
}
   
template<class A, class B>
ostream& operator<<(ostream& out, const Pair<A,B>& pp)
{ 
    out << "{" << pp.first() << "," << pp.second() << "}"; 
    return out;
}

template<class A, class B>
bool  Pair<A,B>::operator==( const Pair<A,B>& y )  const
{
    if( this==&y ) return 1;
      
    return ( first()==y.first() && second()==y.second() );
}
   
// standard product order on pairs
template<class A, class B>
bool  Pair<A,B>::operator<( const Pair<A,B>& T )   const
{
    if( first()<T.first() || (first()==T.first() && second()<T.second()) ) 
	return 1;
    return 0;
}

//---------------------------------------------------------------------------
//  triples
//---------------------------------------------------------------------------

template<class A, class B, class C> class Triple
{
    A  eins;
    B  zwei;
    C  drei;
   
public:

    Triple(): eins(A()), zwei(B()), drei(C()) {}
    Triple( const Triple& src ): eins(src.eins), zwei(src.zwei), drei(src.drei) {}

//     template<class A, class B, class C>
//     Triple( const Triple<A,B,C>& src ): 
// 	eins(src.eins), zwei(src.zwei), drei(src.drei) {}

    Triple( const A& aa, const B& bb, const C& cc ): eins(aa), zwei(bb), drei(cc) {}

    A&        first()        { return eins; } 
    B&        second()       { return zwei; }
    C&        third()        { return drei; }
    const A&  first()  const { return eins; } 
    const B&  second() const { return zwei; }
    const C&  third()  const { return drei; }
    void  get( A& aa, B& bb, C& cc ) const             { aa = eins;bb = zwei;cc = drei; } 
    void  set( const A& aa, const B& bb, const C& cc ) { eins = aa;zwei = bb;drei = cc; } 

    
    Triple&  operator=( const Triple& src );
    bool  operator<( const Triple& src ) const;
    bool  operator==( const Triple& src ) const;

    friend ostream& operator<<<A,B,C>(ostream& out, const Triple& pp);
    
};

//--------------------------------------------------------------------------------
// free function to create triples

template<class A, class B, class C>
Triple<A,B,C>  to_triple( const A& a, const B& b, const C& c )
{
    return  Triple<A,B,C>(a,b,c);
}

//--------------------------------------------------------------------------------

template<class A, class B, class C> 
inline Triple<A,B,C>&  Triple<A,B,C>::operator=( const Triple<A,B,C>& src )
{
    if( this==&src ) return *this;

    eins = src.eins;
    zwei = src.zwei;
    drei = src.drei;

    return *this;
}
   
template<class A, class B, class C>
bool  Triple<A,B,C>::operator<( const Triple<A,B,C>& r ) const
{
    if( first()<r.first() || (first()==r.first() && second()<r.second()) || 
	(first()==r.first() && second()==r.second() && third()<r.third() ) )
	return 1;
    return 0;
}

template<class A, class B, class C> 
bool  Triple<A,B,C>::operator==( const Triple<A,B,C>& r ) const
{
    if( this==&r ) return 1;
      
    return ( first()==r.first() && second()==r.second() && third()==r.third() );
}

template<class A, class B, class C> 
ostream& operator<<(ostream& out, const Triple<A,B,C>& pp)
{ 
    out << "{" << pp.first() << "," << pp.second() << "," << pp.third() << "}"; 
    return out;
}


//----------------------------------------------------------------------
// typedefs, special ops, special sorting
//----------------------------------------------------------------------

typedef   Pair<int,int>          II_Pair;
typedef   Triple<int,int,int>    III_Triple;
typedef   Triple<int,char,int>   ICI_Triple;          // only for output 

//--------------------------------------------------------------------------------
// function objects to manipulate triples

template<class A, class B, class C> struct reverseTriple 
{
    reverseTriple<A,B,C>() {}

    Triple<A,B,C>  operator()( const Triple<A,B,C>& T ) 
    {
	A a;  B b;  C  c;
	T.get(a,b,c);
	return Triple<A,B,C>(c,b,a);
    }
};



// function object for 213 order on triples

template<class A, class B, class C>  struct comp213
{
    comp213<A,B,C>() {}
	
    bool  operator()( const Triple<A,B,C>& S, const Triple<A,B,C>& T ) 
    {
	A a, aa;  B b, bb;  C  c, cc;
	S.get(aa,bb,cc);
	T.get(a,b,c);
	if( bb < b || ( bb==b && aa < a ) || (bb==b && aa==a && cc < c ))
	    return 1;
	return 0;
    }
};


// equality ignoring third component

template<class A, class B, class C>  struct equal12
{
    equal12<A,B,C>() {}
    bool  operator()( const Triple<A,B,C>& S, const Triple<A,B,C>& T ) 
    {
	return  S.first()==T.first() && S.second()==T.second();
    }
};

// order ignoring third component

template<class A, class B, class C> struct  comp12
{
    comp12<A,B,C>() {}
    bool  operator()( const Triple<A,B,C>& S, const Triple<A,B,C>& T ) 
    {
	A a, aa;  B b, bb;  C  c, cc;
	S.get(aa,bb,cc);
	T.get(a,b,c);
	return ( aa<a || ( aa==a && bb<b ) );
    }
};


//----------------------------------------------------------------------
#endif



