//---------------------------------------------------------------------------
//
//  Time-stamp: <2003-11-24 18:43:05 sutner> 
//
//  :File:     ptree.h
//  :Purpose:  parse trees
//  
//---------------------------------------------------------------------------

#ifndef  _PARSETREE_H_
#define  _PARSETREE_H_

#include  "basic.h"
#include  "functoid.h"
#include  "symtab.h"
#include  "urlist.h"
#include  <slist.h>

//------------------------------------------------------------------------------
// classes defined herein
//------------------------------------------------------------------------------

class Expr;			// expressions for parse tree 
class FuncNode;			// user defined function 
class StatNode;			// statements for parse tree 

//------------------------------------------------------------------------------
// StatNode: base class for syntactic constructs 
//    - DefineStat
//    - IfStat
//    - WhileStat
//    - AssgnStat
//    - CompStat
// Expr: base class for expressions
//    - VarExpr     just the name
//    - AssgnExpr   assignments are expressions and statements
//    - UFuncExpr   built-in operations
//    - BFuncExpr   built-in operations
//    - FuncExpr    user defined functions 
//------------------------------------------------------------------------------

class SymTable;

//--------------------------------------------------------------------------------

class VarExpr: public RefAtom
{
    char  *name;

public:
    VarExpr( char *nm );
    ~VarExpr();

    void      print(ostream& out) const;
    UList     eval( SymTable& env );
    AType     atom_type() const { return VEXN; }
};

class AssgnExpr: public RefAtom
{
private:
    char    *name;
    UList   expr;

public:
    AssgnExpr( char *nm, UList ex );
    ~AssgnExpr();

    void      print(ostream& out) const;
    UList     eval( SymTable& env );
    AType     atom_type() const { return AEXN; }
};


class  IfNode:  public RefAtom
{
private:
    UList  cond;
    UList  ifpart;
    UList  elsepart;
    
public:
    IfNode( UList c, UList ifp, UList elp );
    ~IfNode();
    
    void      print(ostream& out) const;
    UList     eval( SymTable& env );
    AType     atom_type() const { return IFN; }
};

class  WhileNode:  public RefAtom
{
private:
    UList  cond;
    UList  body;
    
public:
    WhileNode( UList c, UList bd );
    ~WhileNode();
	
    void      print(ostream& out) const;
    UList    eval( SymTable& env );
    AType     atom_type() const { return WHLN; }
};


class  WhileReadNode:  public RefAtom
{
private:
    char  *vname;
    char  *fname;
    UList  body;
    
public:
    WhileReadNode( const char *vn, const char *fn , UList bd );
    ~WhileReadNode();
	
    void      print(ostream& out) const;
    UList     eval( SymTable& env );
    AType     atom_type() const { return WHLN; }
};


class  CompNode:  public RefAtom
{
private:
    UList  stmlist;
    
public:
    CompNode( UList sl );
    ~CompNode();

    void      print(ostream& out) const;
    UList     eval( SymTable& env );
    AType     atom_type() const { return COMPN; }
};


//--------------------------------------------------------------------------------
// standard operations

class FuncExpr: public RefAtom
{
private:
    char  *fname;
    UList  argl;
    int    alen;
    
public:
    FuncExpr( const char *nm, UList ar, int alen=-1 );

    ~FuncExpr();

    void      print(ostream& out) const;
    UList     eval( SymTable& env );
    AType     atom_type() const { return FEXN; }
};

//--------------------------------------------------------------------------------
// user defined k-ary functions
// UserFunction:   entries in symtab
// UserFuncExpr:   built by parser, evaluation requires proper symtab entry
// DefNode:        built by parser, evaluation creates symtab entry

struct  UserFunction: public RefAtom
{
    size_t      ari;		// arity, error checking 
    char      **namel;		// formal parameter names
    UList       body;		// the function body
    SymTable   *cntxt;		// the context
    
public:
    UserFunction( UList nl, UList bd, SymTable *envp );
    ~UserFunction();

//    UList      eval( SymTable& env );
    void       print(ostream& out) const;
    AType      atom_type() const  { return UFCTN; }
};

struct  UserFuncExpr: public RefAtom
{
    char   *fname;
    UList   argl;
    
public:
    UserFuncExpr( const char* nm, UList al );
    ~UserFuncExpr();

    UList      eval( SymTable& env );
    void       print(ostream& out) const;
    AType      atom_type() const  { return UFCTE; }
};

class  DefNode:  public RefAtom
{
private:
    char    *fname;		// function name 
    UList   paral;		// parameter names
    UList   body;		// function body
    
public:

    DefNode( char *fn, UList nl, const UList bd );
    ~DefNode();
    
    void      print(ostream& out) const;
    UList     eval( SymTable& env );
    AType     atom_type() const { return DEFN; }
};

//--------------------------------------------------------------------------------
#endif

//  Local Variables: ***
//  eval: (setq case-fold-search nil) ***
//  eval: (setq case-replace nil) ***
//  End: ***
