//--------------------------------------------------------------------------------
//
//  Time-stamp: <2003-11-24 18:46:30 sutner>
// 
//  :File:     sgroup.h 
//  :Purpose:  semigroup generation, semimodule generation
//             dump n-ary tries in general case, catastrophic for large n
//             
//--------------------------------------------------------------------------------

#ifndef _SGROUP_H_
#define _SGROUP_H_

//----------------------------------------------------------------------

#include  <hash_map.h>
#include  <hash_set.h>
#include  <iostream>
#include  <list>

#include  "basic.h"
#include  "fsm.h"
#include  "graph.h"
#include  "intset.h"
#include  "ioatom.h"
#include  "pairs.h"
#include  "trees.h"

//----------------------------------------------------------------------
// defined herein

class  SGroup;
class  DClass;

typedef   bTrieNode   *bTrieP;
typedef   nTrieNode   *nTrieP;
typedef   Pair<string,string>  Equation;      // switch to char* ???
typedef   Pair<nTrieP,string>                 SGelem;

typedef   hash_set<uvector, hash<uvector>, equal_to<uvector> >   KISet;
typedef   hash_map<uvector, ilist, hash<uvector>, equal_to<uvector> >   HclassMap;


//--------------------------------------------------------------------
//  semigroups:
//  H-classes, D-classes, semigroups
//--------------------------------------------------------------------

class  DClass
{
    size_t   kdim;			// number of L-classes (kernels)
    size_t   idim;			// number of R-classes (images)
    size_t   hdim;			// size if an H-class
    size_t   wdth;			// size of functions
    size_t   wh;			// speed up read/write
    size_t   whi;			// 
    uvector   arr;		// data array cum access

public:

    DClass() {}
    DClass( size_t k, size_t i, size_t h, size_t w ):
	kdim(k), idim(i), hdim(h), wdth(w), wh(w*h), whi(w*h*i), arr(w*h*i*k) {} 

    ~DClass();

    size_t     kSize() const {return kdim;}
    size_t     iSize() const {return idim;}
    size_t     hSize() const {return hdim;}
    size_t     width() const {return wdth;}
    size_t      size() const {return arr.size();}
    uvector&  data()     {return arr;}

    III_Triple  dimensions() const 
                {III_Triple dim(kdim,idim,hdim); return dim;}

    void  initialize( size_t k, size_t i, size_t h, size_t n );

    size_t  get( size_t k, size_t i, size_t h, size_t j ) const 
    {
	assert( 0 <= k && k < kdim );
	assert( 0 <= i && i < idim );
	assert( 0 <= h && h < hdim );
	assert( 0 <= j && j < wdth );
	return  arr[ j + h*wdth + i*wh + k*whi ];
    }

    DClass&   operator=( const DClass& src );    //needed ???  
    bool      operator==( const DClass& src );   //needed ???

    void   print_dclass( ostream& ) const;

};

//----------------------------------------------------------------------

class  SGroup
{
    size_t     num;		// functions [num] --> [num] 
    size_t     len;		// size of semigroup 
    size_t     oneflg;		// is id element of semigroup 
    size_t     numid;		// number of idempotents

    uvector          idcnt;	// number of idempotents by rank
    uvector            sgr;	// flat array of sg elements 
    vector<string>       wit;	// witnesses 
    list<Equation>       equ;	// equations 
    vector<III_Triple> dcldim;	// D-class dimensions 
    vector<DClass>       dcl;	// D-classes 


public:

    SGroup(): num(0),len(0),oneflg(0),numid(0) {}
    explicit SGroup( size_t n ): num(n),len(0),oneflg(0),numid(0), idcnt(n+1) {}
    
    SGroup( const SGroup& src );

    ~SGroup();
    
    SGroup&  operator=( const SGroup& src );
   
    size_t&   width()                { return num; }
    size_t    width() const          { return num; }
    size_t&   size()                 { return len; }
    size_t    size() const           { return len; }
    size_t&   oneflag()              { return oneflg; }
    const  size_t oneflag() const    { return oneflg; }
    size_t&   numIdempotents()       { return numid; }
    size_t    numIdempotents() const { return numid; }
    size_t    numEquations() const   { return equ.size(); }
    size_t    numDClasses() const    { return dcl.size(); }
   
    uvector&             numIdemRank()       { return idcnt; }
    const uvector        numIdemRank() const { return idcnt; }
    uvector&             semigroup()         { return sgr; }
    const uvector&       semigroup() const   { return sgr; }
    vector<string>&        witnesses()         { return wit; }
    const vector<string>&  witnesses() const   { return wit; }
    list<Equation>&        equations()         { return equ; }
    const list<Equation>&  equations() const   { return equ; }
    vector<DClass>&        dclasses()          { return dcl; }
    const vector<DClass>&  dclasses() const    { return dcl; }

    vector<III_Triple>&    dclassDimensions()  { return dcldim; }
    const vector<III_Triple>&  dclassDimensions() const { return dcldim; }

    void   simplify_equations();

    bool   aperiodic() const;

    void   print_elements( ostream& os ) const;
    void   print_witnesses( ostream& os ) const;
    void   print_equations( ostream& os ) const;
    void   print_dclasses( ostream& os ) const;

    friend ostream& operator<<(ostream& out, const SGroup& sg );   
};

//--------------------------------------------------------------------------------
//  PSemiModuleGen
//  generate sub-semimodule of pow(Q)
//  -- binary trie, leaves carry counter
//  -- linked list of pairs of pointers to a leaf in binary trie
//  -- delta:  (m)x(k)  flattened matrix, transition function of power automaton
//--------------------------------------------------------------------------------

class  PSemiModuleGen
{
private:
    int                  n;
    int                  k;
    bTrie                btrie;         // the binary trie
    list<bTrieP>         el;		// list of pointers to leaves in binary trie
    vector<ilist>        op;		// generating function, n by k matrix of sets 
    ulist                del;		// transition function for result
    
public:
    typedef  list<bTrieP>::iterator     iterator;

    PSemiModuleGen(){}
    PSemiModuleGen( int nn, int kk ): n(nn), k(kk), btrie(n), el(), op(n*k), del() {}
    ~PSemiModuleGen();

    size_t             size()        { return el.size(); }
    vector<ilist>&     next()        { return op; }
    list<bTrieP>&      elements()    { return el; }
    ulist&             delta()       { return del; }

    iterator  begin()  { return  el.begin(); }
    iterator  end()    { return  el.end(); }
   
    void  current_set( iterator itr, IntSet& st );
    void  insert( const IntSet &ini );
    int   generate();

};				// end class PSemiModuleGen

    
   

//----------------------------------------------------------------------
//  generate transformation semigroups
//  -- n-ary trie
//  -- linked list of pairs of pointers (p,q)
//     p points to a leaf in n-ary trie, q a witness 
//----------------------------------------------------------------------

class  SGroupGen
{
public:

    typedef  vector<SGelem>::iterator   iterator;	  
    enum  GenType { GENW, GENE, GEND };
    

private:
    size_t                 wid;           // width
    size_t                 numgen;	  // number of generators  
    size_t                 oneflg;        // is monoid
    uvector                gens;          // generators, flat 
    
    nTrie                  ntrie;	  // n-ary trie for functions [n] --> [n]
    HclassMap              hclasses;      // elms ordered by kerim  

    vector<SGelem>         el;		  // list of elements/witnesses
    list<III_Triple>       eqs;		  // rw-rules in form (wit1,sym,wit2)
    vector<ilist>          idarray;       // idempotents by rank
    list<ilist>            complist;      // compatible idempotents
    

public:

    SGroupGen() {}
    SGroupGen( size_t w, size_t k, const uvector& gg );
    ~SGroupGen();

    void  generate( SGroup& sg, GenType how = GENW );

    size_t    size()           { return ntrie.size(); }
    size_t    width()          { return wid; }
    size_t    numGenerators()  { return numgen; }
    uvector&  generators()     { return gens; }

    iterator  begin()          { return  el.begin(); }
    iterator  end()            { return  el.end(); }

    void  current( iterator itr, uvector &x ) { retrieve_branch( itr->first(), x ); }
    void  current( size_t i, uvector &x )
    {
	retrieve_branch( el[i].first(), x );
    }

    // write data out to a SGroup
    void  write_semigroup( SGroup& sg );
    void  write_equations( SGroup& sg );
    void  write_dclasses( SGroup& sg );

    
private:

    // auxiliary functions 
    size_t  split_idem_lists(); 
};				// end class SGroupGen


//--------------------------------------------------------------------
#endif

