//---------------------------------------------------------------------------
//
//  Time-stamp: <2003-11-24 18:21:03 sutner> 
//
// :File:      ioatom.cc
// :Purpose:   I/O for atoms, in particular finite state machines
//             reading really should use a little parser
//
//             two formats:
//             -  files
//             -  aut 
//---------------------------------------------------------------------------

#include  <list>
#include  <string>
#include  <vector>

#include  "fsm.h"
#include  "ioatom.h"
#include  "listutil.h"
#include  "stl-wrap.h"
#include  "urlist.h"
#include  "useratom.h"
#include  "util.h"

//--------------------------------------------------------------------------------
//  File I/O
//
//  currently only read  SA, FA, DFA, CA
//  performs only minimal error checking, should really be a little parser
//  format:
//       IDENT
//       size info
//       actual data
//  
//--------------------------------------------------------------------------------


UList   read_atom_raw( istream& in )
{
    string  x;
    int  p, s, q, n, k, ni, nf;
    
    in >> x;
    in >> n >> k;
    
    if( x == "SA" )
    {
	bool  eps = 0;
	UList  res = new SAAtom(n,k);
	SA   &sa = savalue(res);
	while( in >> p >> s >> q )
	{
	    assert( 0 <= s && s <= k );
	    assert( 0 < p && p <= n );
	    assert( 0 < q && q <= n );
	    if( !s ) eps = 1;
	    sa.transitionsInsert( p-1, s-1, q-1 );
	}
	sa.eps_free() = eps ? FALSE : TRUE;
	return  res;
    }
    
    if( x == "FA" )
    {
	bool   eps = 0;
	UList  res = new FAAtom(n,k);
	FA    &fa  = favalue(res);
	in >> ni >> nf;
	for( p = 0; p < ni; ++p )
	{
	    in >> q;
	    assert( 0 < q && q <= n );
	    fa.initial().insert( q-1 );
	}
	for( p = 0; p < nf; ++p )
	{
	    in >> q;
	    assert( 0 < q && q <= n );
	    fa.final().insert( q-1 );
	}
	while( in >> p >> s >> q )
	{
	    assert( 0 <= s && s <= k );
	    assert( 0 < p && p <= n );
	    assert( 0 < q && q <= n );
	    if( !s ) eps = 1;
	    fa.transitionsInsert( p-1, s-1, q-1 );
	}
	fa.eps_free() = eps ? FALSE : TRUE;
	return  res;
    }
    
    if( x == "DFA" )
    {
	UList  res = new DFAAtom( n, k );
	DFA  &dfa = dfavalue( res );
	in >> q >> nf;
	dfa.initial() = q-1;
	for( p = 0; p < nf; ++p )
	{
	    in >> q;
	    assert( 0 < q && q <= n );
	    dfa.final().insert( q-1 );
	}
	for( s = 0; s < k; ++s )
	    for( p = 0; p < n; ++p )
	    {
		in >> q;
		assert( 0 < q && q <= n );
		dfa.transitions()[ n*s + p ] = q-1;
	    }
	return  res;
    }

    if( x == "CA" )
    {
	UList  res = new CAAtom();
	CA   &ca = cavalue(res);
	ca.width() = n;
	ca.numSymbols() = k;
	in >> x;
	ca.rule().set( x.c_str(), 10 );
	return  res;
    }
    
    return nil;
}


inline   void  print_trans( ostream& os, const Trans& t, int off=1 )
{
    os << setw(4) << (t.p+off);
    os << setw(4) << (t.s+off);
    os << setw(4) << (t.q+off) << endl;
}


//--------------------------------------------------------------------------------

void   write_atom_raw( ostream& os, const SA& sa, bool norm )
{
    int  ww = (sa.numStates()>999) ? 8 : 4;

	os << "SA" << endl;
	os << setw(ww) << sa.numStates() << setw(ww) << sa.numSymbols() << endl;

	TransV::const_iterator  it;
	foreach_item( it, sa.transitions() )
	    print_trans( os, *it );

	if( !norm && sa.states() )
	    os << sa.states() << endl;
}


void   write_atom_raw( ostream& os, const FA& fa, bool norm )
{
    int  ww = (fa.numStates()>999) ? 8 : 4;

	os << "FA" << endl;
	os << setw(ww) << fa.numStates() << setw(ww) << fa.numSymbols() << endl;
	os << setw(ww) << fa.initial().size() << setw(ww) << fa.final().size() << endl;
	fa.initial().print( os, 1 ); 
	fa.final().print( os, 1 );
	TransV::const_iterator  it;
	foreach_item( it, fa.transitions() )
	    print_trans( os, *it );

	if( !norm && fa.states() )
	    os << fa.states() << endl;
}


void   write_atom_raw( ostream& os, const DFA& dfa, bool norm )
{
    int  ww = (dfa.numStates()>999) ? 8 : 4;

	os << "DFA" << endl;
	int n = dfa.numStates();
	int k = dfa.numSymbols();
	os << setw(ww) << n << setw(ww) << k << endl;
	os << setw(ww) << (dfa.initial()+1) << setw(ww) << dfa.final().size() << endl;
	dfa.final().print( os, 1 ); 

	const vector<size_t>  &mat = dfa.transitions();
	for( int i = 0; i < k; ++i )
	{
	    for( int j = 0; j < n; ++j )
		os << setw(ww) << (mat[ n * i + j ] + 1); 
	    os << endl;
	}
	if( !norm && dfa.states() )
	    os << dfa.states() << endl;
}

void   write_atom_raw( ostream& os, const CA& ca, bool norm )
{
    int  ww = 4;

	os << "CA" << endl;
	os << setw(ww) << ca.width() << setw(ww) << ca.numSymbols() << endl;
	os << ca.rule() << endl;
}


void   write_atom_raw( ostream& os, const SGroup& sg, bool norm )
{
    int  ww = (sg.size()>999) ? 8 : 4;

	os << "SGroup" << endl;
	os << setw(ww) << sg.size() << setw(ww) << sg.width();
	os << setw(ww) << sg.oneflag() << endl;
	os << setw(ww) << sg.numIdempotents() << endl;
	os << sg.numIdemRank() << endl;
	sg.print_elements( os );
	os << endl;
	sg.print_equations( os );
	os << endl;
	sg.print_dclasses( os );
}


       
void   write_atom_raw( ostream& os, UList A, bool norm )
{
    switch( A.atom_type() )
    {
    case SAA:  write_atom_raw( os, savalue( A ), norm ); break;
    case FAA:  write_atom_raw( os, favalue( A ), norm ); break;
    case DFAA: write_atom_raw( os, dfavalue( A ), norm ); break;
    case CAA:  write_atom_raw( os, cavalue( A ), norm ); break;
    default:   break;
    }
}



//--------------------------------------------------------------------------------
// cooked 

UList  read_atom( UList str )
{
    ifstream  inst( svalue(str)  );
    if( !inst )
    {
	cerr << "read_atom:  cannot open " << svalue(str) << endl;
	return nil;
    }
    
    return read_atom_raw( inst );
}

       
UList  write_atom( UList A, UList S )
{
    string  fname;
    UList   st, no;
    int norm = 1;
    
    if( S.atom_type() == STR )
	fname = svalue( S );
    else
    {
	S >> st >> no;
	fname = svalue( st );
	norm = ivalue( no );
    }
    
    ofstream os( fname.c_str() );
    if( !os )
    {
	cerr << "write_atom:  cannot open " << fname << endl;
	return nil;
    }
    
    switch( A.atom_type() )
    {
    case SAA:  write_atom_raw( os, savalue( A ), norm );  break;
    case FAA:  write_atom_raw( os, favalue( A ), norm );  break;
    case DFAA: write_atom_raw( os, dfavalue( A ), norm ); break;
    case CAA:  write_atom_raw( os, cavalue( A ), norm );  break;

    default:   os << A << endl; break;
    }
    os.close();
    return nil;
}

    
//--------------------------------------------------------------------------------
//  I/O to screen 
//--------------------------------------------------------------------------------


ostream& operator<<( ostream& out, const SA& M )
{
   out << "SA[ " << M.numStates() << COMMA << M.numSymbols() << COMMA; 
   ::PrintTransitions( out, M.transitions(), 1 );
   out << "]";
   return out;
}

ostream& operator<<( ostream& out, const FA& M )
{
   out << "FA[ " << M.numStates() << COMMA << M.numSymbols() << COMMA; 
   ::PrintTransitions( out, M.transitions(), 1 );
   out << COMMA;
   Print( out, M.initial(), 1 ); 
   out << COMMA;
   Print( out, M.final(), 1 ); 
   out << "]";
   return out;
}

ostream& operator<<( ostream& out, const DFA& M )
{
   out << "DFA[ " << M.numStates() << COMMA << M.numSymbols() << COMMA; 
   const vector<size_t>  &tr = M.transitions();
   ::PrintMatrixManip( out, tr.begin(), tr.size(),
		       bind2nd(plus<int>(),1), M.numStates() );
   out << COMMA << M.initial() + 1;
   out << COMMA;
   Print( out, M.final(), 1 ); 
   out << "]";
   return out;
}

//---------------------------------------------------------------------


