//---------------------------------------------------------------------------
//
//  Time-stamp: <2003-11-24 18:24:31 sutner> 
//
//  :File:     parsetree.cc
//  :Purpose:  atom types for parse tree implementation
// 
//---------------------------------------------------------------------------

#include  <iostream>
#include  <fstream>
#include  <sstream>

#include  "basic.h"
#include  "ptree.h"
#include  "urlist.h"
#include  "useratom.h"
#include  "listutil.h"
#include  "symtab.h"

//--------------------------------------------------------------------------------
// Expressions
//--------------------------------------------------------------------------------


VarExpr::VarExpr( char *nm ): name(NULL)
{
    write_string( name, nm );
}
VarExpr::~VarExpr() { delete [] name; }

void   VarExpr::print(ostream& os) const { os << "VarExpr"; }

UList  VarExpr::eval( SymTable& env ) 
{ 
    EVALMSSG("var atom");
    return env.lookup( name ); 
}

//--------------------------------------------------------------------------------

AssgnExpr::AssgnExpr( char *nm, UList ex ): name(NULL), expr(ex)
{
    write_string( name, nm );
}

AssgnExpr::~AssgnExpr() { delete [] name; }

void AssgnExpr::print(ostream& os) const { os << "AssgnExpr  " << name << endl; }

UList  AssgnExpr::eval( SymTable& env )
{
    EVALMSSG("assg atom");

    UList val = expr.eval( env );
    env.assign( val, name );
    return  val;
}

//--------------------------------------------------------------------------------
// Statements
//--------------------------------------------------------------------------------

IfNode::~IfNode() {}
IfNode::IfNode( UList c, UList ifp, UList elp ): 
    cond(c), ifpart(ifp), elsepart(elp) {}
    
void       IfNode::print(ostream& os) const { os << "IfNode"; }

UList     IfNode::eval( SymTable& env )
{
    EVALMSSG("if atom");

    if( isTrue( cond.eval(env) ) )
	return ifpart.eval( env );
    else
	return elsepart.eval( env );
}


//--------------------------------------------------------------------------------

WhileNode::~WhileNode() {}

WhileNode::WhileNode( UList c, UList bd ): cond(c), body(bd) {}
    
void       WhileNode::print(ostream& os) const { os << "WhileNode"; }

UList  WhileNode::eval( SymTable& env )
{
    EVALMSSG("while atom");

    UList res;
    
    while( isTrue( cond.eval(env) ) )
	res = body.eval( env );

    return res;
}

//--------------------------------------------------------------------------------

WhileReadNode::~WhileReadNode() 
{
    delete  fname;
    delete  vname;
}

WhileReadNode::WhileReadNode( const char *vn, const char *fn , UList bd ): body(bd) 
{
    write_string( vname, vn );
    write_string( fname, fn );
}

void       WhileReadNode::print(ostream& os) const { os << "WhileReadNode"; }

UList  WhileReadNode::eval( SymTable& env )
{
    ifstream  inf(fname);
    if( !inf )
    {
	cerr << "attach: cannot open input stream " << fname << endl;
	return nil;
    }

    extern  Parser  *_einmal_;

    UList  res;
    char  buffer[1024];		// will overflow some day
    char  *bp;
    char  c;
    while( inf.get(c) )
    {
	bp = buffer;
	*bp++ = c;
	while( (c = inf.get()) && c != '\n' ) *bp++ = c;
	*bp++ = '\n';
	*bp++ = '\0';

	_einmal_->attach( "string", buffer );
	UList val = _einmal_->parse( env );
	env.assign( val, vname );
	res = body.eval( env );
    }
    
    return res;
}


//--------------------------------------------------------------------------------

CompNode::~CompNode() {}

CompNode::CompNode( UList sl ): stmlist(sl) {}
    
void       CompNode::print(ostream& os) const { os << "CompNode"; }

UList  CompNode::eval( SymTable& env )
{
    EVALMSSG("comp atom");

    UList  x, res, st  = stmlist; 
    while( st )
    {
	st >> x;
	res = x.eval( env );
    }
    return res;
}

//--------------------------------------------------------------------------------
//  alen: computed by parser
//  use for error checking, default arguments

FuncExpr::FuncExpr( const char *nm, UList ar, int n ): 
    fname(NULL), argl(ar), alen(n)
{
    write_string( fname, nm );
}

FuncExpr::~FuncExpr() { delete [] fname; }

void FuncExpr::print(ostream& os) const 
{ 
    os << "FuncExpr" << fname << "  " << argl;
}

static UList  eval_args( UList args, SymTable& env )
{
    UList  h, vals;
    while( args )
    {
        args >> h;
        vals += h.eval( env );
    }
    return  reverse(vals);
}

// #args mismatch handling a bit wild
// allows for  'CA[a,b,c]' as well as 'CA[(a,b,c)]'

UList  FuncExpr::eval( SymTable& env )
{
    extern SymTable *_glob_op_tab_;
    
				//  get corresponding Function 
    UList  F = _glob_op_tab_->lookup( fname ); 

    if( !F ) return nil;

    int  ari = arity( F );
				// get rid of nullary
    if( ari == 0 ) return F();

    UList  vargl;
				// special case: map, argl[0] is oper name
    if( strcmp(fname,"map")==0 )
    {
	UList  G = _glob_op_tab_->lookup( svalue(*argl) );
	++argl;
	return  F( G, (*argl).eval( env ) );
    }
    else
	vargl = eval_args( argl, env );

    
//    cerr << "*** " << fname << " : " << ari << " : " << alen << "   " << vargl << endl;


				// too many arguments --> one list
    if( alen > ari ) return F( vargl );

				// must have 1 - 3 arguments 
    switch( ari )
    {
    case 1:    
	return F( *vargl );
	break;

    case 2:
	if( alen == 1 ) 
	    return F( *vargl, nil );
	else
	    return F( *vargl, *(vargl.tail()) );
	break;

    case 3:    
	if( alen == 2 )
	    return F( *vargl, *(vargl.tail()), nil );
	else
	    return F( *vargl, *(vargl.tail()), *(vargl.tail().tail()) );
	break;
	

    default:
	return nil;
    }
}

//--------------------------------------------------------------------------------

UserFunction::UserFunction( UList nl, const UList bd, SymTable *envp ): 
    body(bd), cntxt(envp)
{
    ari = length_int( nl );
    namel = new (char*)[ari];

    int i = 0;
    UList S;
    while( nl )
    {
	nl >> S;
	namel[i] = new  char[strlen(svalue(S))+1];
	strcpy( namel[i], svalue(S) );
	i++;
    }
}

UserFunction::~UserFunction() 
{ 
    for( size_t i = 0; i < ari; i++ )
	delete [] namel[i]; 
    delete [] namel;
}

void UserFunction::print(ostream& os) const { os << "UserFunction"; }

//--------------------------------------------------------------------------------

UserFuncExpr::UserFuncExpr( const char* nm, UList al ): fname(NULL), argl(al)
{
    write_string( fname, nm );
}

UserFuncExpr::~UserFuncExpr() { delete fname; }

void UserFuncExpr::print(ostream& os) const { os << "UserFuncExpr"; }


// ugly use of dynamic_cast, fix

UList  UserFuncExpr::eval( SymTable& env )
{
    UList  F = env.lookup( fname );
    UserFunction  *fp = dynamic_cast<UserFunction*>( ufunc( F ) );
    if( !fp ) return nil;    

    UList  vargs = eval_args( argl, env );

    SymTable  parenv( fp->cntxt );
    parenv.assign_list( vargs, fp->namel );

    return  fp->body.eval( parenv );
}

//--------------------------------------------------------------------------------


DefNode::DefNode( char *fn, UList nl, const UList bd ): 
    fname(NULL), paral(nl), body(bd)
{
    write_string( fname, fn );
}

DefNode::~DefNode() { delete [] fname; }

void DefNode::print(ostream& os) const { os << "DefNode: " << fname << endl; }

UList    DefNode::eval( SymTable& env )
{
    EVALMSSG("defun atom");

    UList  fn = new UserFunction( paral, body, &env );
    env.assign( fn, fname );
    return new StringAtom(fname);
}


AtomP  ufunc( const UList& x ) { return x.ap; }


//--------------------------------------------------------------------------------

//  Local Variables: ***
//  eval: (setq case-fold-search nil) ***
//  eval: (setq case-replace nil) ***
//  End: ***
