//--------------------------------------------------------------------------------
//
//  Time-stamp: <2003-11-24 18:29:29 sutner>
// 
//  :File:     sgroup-new.cc
//  :Purpose:  Generating syntactic semigroup, D-class decomposition. 
//            
//--------------------------------------------------------------------------------

#include  <iostream>
#include  <iomanip>
#include  <stack>

#include  "basic.h"
#include  "sgroup.h"
#include  "set.h"

// #define  SGDEBUG

//----------------------------------------------------------------------
// helpers 

static  int   simplify_rw( list<Equation>& eq );

static inline  void  cross_over( const uvector& K1, const uvector& I1, 
			  const uvector& K2, const uvector& I2, 
			  uvector& AB, uvector& BA )
{
    int n = K1.size();
    for( int i = 0; i < n; i++ )
    {
	AB[i] = K1[i]; AB[n+i] = I2[i]; 
	BA[i] = K2[i]; BA[n+i] = I1[i];
    }
}

static void kernel_image( const uvector& x, uvector& ker, uvector& im );

//--------------------------------------------------------------------
//  generate SGroup from Generators
//  how:   GENW  - only semigroup and witnesses
//         GENE  - plus equations
//         GEND  - plus D-class decomposition
//--------------------------------------------------------------------

SGroupGen::~SGroupGen() {}


// write generators, check for equality

SGroupGen::SGroupGen( size_t w, size_t k, const uvector& gg ):
    wid(w), numgen(k), oneflg(0), gens(gg), ntrie(wid,wid), idarray(wid+1)
{
    string    wit;
    uvector   kerim(2*wid);
    uvector   aux(wid);
    nTrieP    leaf;
    pair<HclassMap::iterator,bool>  itbp;

    for( size_t s = 0; s < numgen; s++ )
    {
	register size_t  i, z; 
	bool   isidem = 1;
	size_t rk = 0; 

	Fill( kerim, 0 );
   
	for( i = 0; i < wid; i++ )
	{
	    z = gens[wid*s + i];
	    if( kerim[ wid + z ] )		
		kerim[i] = aux[z];
	    else
	    {
		kerim[i] = aux[z] = i;
		kerim[ wid + z ] = 1;		    
		rk++;
		if( isidem && gens[ wid*s + z ] != z )
		    isidem = 0;
	    }
	}     

	if( isidem && rk == wid ) oneflg = 1;

	bool  oldel = ntrie.probe_branch( gens.begin() + s*wid, leaf );

	if( !oldel )	// element new
	{
	    if( isidem )
		idarray[rk].push_back( size()-1 );

	    wit = char( 'a' + s );
	    el.push_back( SGelem( leaf, wit ) );

	    itbp = hclasses.insert( HclassMap::value_type (kerim,ilist()) );
	    (*itbp.first).second.push_back(size()-1);
	}
	else			// element old 
	    eqs.push_back( III_Triple( -1, s, leaf->content() ));

    } // end s loop
}
    

//--------------------------------------------------------------------------------
// generating a semigroup 

void  SGroupGen::generate( SGroup& sg, GenType how )
{
    size_t  rk, s, isidem;
    size_t  n = width();
    size_t  k = numGenerators();

    uvector   kerim(2*n);
    uvector   aux(n);
    uvector   curr(n);
    uvector   next(n);
    pair<HclassMap::iterator,bool>  itbp;
	      

				// main loop:  traverse and build elemlist
    size_t  pp = 0;
    while( pp < el.size() )
    {
	current( pp, curr );	// get current element
				// multiply by all generators
	for( s = 0; s < k; s++ )
	{
	    register size_t  i, z; 
	    isidem = 1;
	    rk = 0; 
	    Fill( kerim, 0 );
   
	    for( i = 0; i < n; i++ )
	    {
		z = next[i] = gens[n*s + curr[i]];
		if( kerim[n+z] )
		    kerim[i] = aux[z];
		else
		{
		    kerim[i] = aux[z] = i;
		    kerim[n+z] = 1;		    
		    rk++;
		    if( isidem && gens[n*s + curr[z]] != z )
			isidem = 0;
		}
	    }     

#ifdef SGDEBUG
	    cout << "next: " << next << endl;
#endif						  
				// check for identity function, useless ???
	    if( isidem && rk==n ) oneflg = 1;

	    nTrieP  leaf;
	    bool  oldel = ntrie.probe_branch( next.begin(), leaf );

	    if( !oldel )	// element new
	    {
		if( isidem )
		    idarray[rk].push_back( size()-1 );

		string wit = el[pp].second() + char( 'a' + s );
		el.push_back( SGelem( leaf, wit ) );

		if( how == GEND ) // enter into kerim map 
		{
		    itbp = hclasses.insert( HclassMap::value_type (kerim,ilist()) );
		    (*itbp.first).second.push_back(size()-1);
		}
	    }
	    else			// element old 
		if( how != GENW ) 
		    eqs.push_back( III_Triple( pp, s, leaf->content() ));

	} // end s loop

	pp++;

    } // end main loop 

#ifdef SGDEBUG
    HclassMap::iterator  kiitr;
    foreach_item( kiitr, hclasses )
	cout << "kerim: " << kiitr->first << " --> " << kiitr->second << endl;
#endif						  

				// write elements and witnesses to sg
    write_semigroup( sg );

    if( how == GENE || how == GEND )
    {
	write_equations( sg );
	simplify_rw( sg.equations() );
    }
    
    if( how == GEND )
    {
	split_idem_lists();
	write_dclasses( sg );
    }

    
}

//--------------------------------------------------------------------
// split idempotent lists in idarray[rk] into compatible blocks,
// store these in master list  complist
// return number of compatible blocks == complist.size()

size_t   SGroupGen::split_idem_lists()
{
    size_t  rk, e, ee;
    ilist::iterator   it, tit;
    
    ilist     L;
    uvector   A(wid), B(wid), K1(wid), K2(wid), I1(wid), I2(wid);
    uvector   AB(2*wid), BA(2*wid);

    for( rk = 1; rk <= wid; ++rk )
	while( !idarray[rk].empty() )       // traverse idem list
	{
	    e = idarray[rk].front();
	    idarray[rk].pop_front();
	    
#ifdef  SGDEBUG
	    cout << "comp base: " << e << endl;
#endif

	    retrieve_branch( el[e].first(), A );

	    kernel_image( A, K1, I1 );
	   
	    complist.push_back( L );          // start new IdemList
	    complist.back().push_back(e);
	    
	    // compare to remaining idems of rank rk
	    for( it = idarray[rk].begin(); it != idarray[rk].end();  )
	    {
		ee = *it;
#ifdef  SGDEBUG
cerr << "--         " << ee << endl;
#endif
		retrieve_branch( el[ee].first(), B );
		kernel_image( B, K2, I2 );
		cross_over( K1, I1, K2, I2, AB, BA );

		if( hclasses.find( AB ) != hclasses.end() && 
		    hclasses.find( BA ) != hclasses.end() )
		{
#ifdef  SGDEBUG
cerr << "--         " << AB << endl;
cerr << "--         " << BA << endl;
#endif
		    complist.back().push_back(ee);
		    tit = it;
		    ++tit;
		    idarray[rk].erase( it );
		    it = tit;
		}
		else
		    ++it;
	    }
#ifdef  SGDEBUG
cerr << "comps: " << complist.back() <<endl;
#endif

	}

   return complist.size();
}


//--------------------------------------------------------------------
//  write semigroup elements and witnesses to SGroup

void  SGroupGen::write_semigroup( SGroup& sg )
{
    int n = sg.width() = width();
    int m = sg.size()  = size();
    sg.oneflag() = oneflg;

    sg.numIdemRank().resize( n );
    sg.semigroup().resize( n*m );
    sg.witnesses().resize( m );
   
    uvector  curr(n);

    // copy ntrie --> sg
    int i = 0;
    SGroupGen::iterator itr;
    uvector::iterator  sgitr = sg.semigroup().begin();

    foreach_item( itr, el )
    {
	current( itr, curr );	// get current element
	copy( curr.begin(), curr.end(), sgitr );
	sgitr += n;
	sg.witnesses()[i++] = itr->second();
    }

				// idarray is 0:n, must shift

    vector<ilist>::const_iterator  it = idarray.begin();
    i = 0;
    size_t  cnt = 0;
    for( ++it; it != idarray.end(); ++it )
	cnt += ( sg.numIdemRank()[i++] = it->size() );
    
    sg.numIdempotents() = cnt; 

}

//--------------------------------------------------------------------
//  write equations to SGroup

void  SGroupGen::write_equations( SGroup& sg )
{
    int  p, s, q;
    list<III_Triple>::const_iterator itr;

    foreach_item( itr, eqs )
    {
	itr->get( p, s, q );
	if( p == -1 )
	    sg.equations().push_back( 
		Equation( string(1,char(s+'a')), el[q].second() ) );
	else
	    sg.equations().push_back( 
		Equation( el[p].second() + char(s+'a'), el[q].second() ) );
    }
}

//--------------------------------------------------------------------
//  compute and write i-th D-class to SGroup

static void kernel_image( const uvector& x, uvector& ker, uvector& im )
{
    size_t i, z, n = x.size();
    
    uvector  aux(n);
    Fill( aux, 0 );
    
    Fill( ker, 0 );
    Fill( im, 0 );
    
    for( i = 0; i < n; i++ )
    {
	z = x[i];
	if( im[z] )		        // z is in the image
	    ker[i] = aux[z];
	else
	{
	    im[z] = 1;
	    ker[i] = aux[z] = i;
	}
    }     
}

void  SGroupGen::write_dclasses( SGroup& sg )
{
    size_t  ith = 0, n = wid;
    size_t  ndcl = complist.size();
    
    uvector     id(n);
    uvector     ker(n);
    uvector     im(n);
    uvector     kerim(2*n);
    KISet         Ker, Im;

    list<ilist >::iterator  itr;
    ilist::iterator   it;
    
				// make lists of all kernels and images
    sg.dclasses().resize( ndcl );
    sg.dclassDimensions().resize( ndcl );

    foreach_item(itr,complist)
    {

	Ker.erase( Ker.begin(), Ker.end() );
	Im.erase( Im.begin(), Im.end() );
	
	foreach_item(it,*itr)	// for each idempotent id
        {
	    current( *it, id );
	    kernel_image( id, ker, im );
	    Ker.insert( ker );
	    Im.insert( im );
	}
   
	copy( ker.begin(), ker.end(), kerim.begin() );
	copy( im.begin(), im.end(), kerim.begin()+n );
	
	int kdim = Ker.size();
	int idim = Im.size(); 
	int hdim = hclasses[kerim].size();

				// initialize i-th D-class in sg
	sg.dclassDimensions()[ith].set(kdim,idim,hdim);
	sg.dclasses()[ith].initialize( kdim, idim, hdim, n ); 
	
	KISet::iterator   kitr;
	KISet::iterator   iitr;
	ilist::iterator       hit;
	
	uvector::iterator   dcitr = sg.dclasses()[ith].data().begin();
   
	foreach_item(kitr,Ker)
        foreach_item(iitr,Im)
	{
	    copy( (*kitr).begin(), (*kitr).end(), kerim.begin() );
	    copy( (*iitr).begin(), (*iitr).end(), kerim.begin()+n );
	
	    foreach_item( hit, hclasses[kerim] )
	    {
		current( *hit, id );
		copy( id.begin(), id.end(), dcitr );
		dcitr += n;
	    }
	}
	ith++;
    }

#ifdef  SGDEBUG
    cout << "dimensions: " << sg.dclassDimensions() << endl;
    for( size_t i=0; i<ndcl; i++ )
	cout << sg.dclasses()[i].data() << endl;
#endif

}

//--------------------------------------------------------------------
//  SGroup, D-classes, ... 
//--------------------------------------------------------------------

SGroup::~SGroup() {}

SGroup::SGroup( const SGroup& src ):
    num(src.num), len(src.len), oneflg(src.oneflg), numid(src.numid)
{
    idcnt = src.idcnt;
    sgr = src.sgr;
    wit = src.wit;
    equ = src.equ;
    dcldim = src.dcldim;
    dcl = src.dcl;
}

SGroup&  SGroup::operator=( const SGroup& src )
{
   if( &src==this )
     return *this;

    num = src.num;    
    len = src.len;
    oneflg = src.oneflg;
    numid = src.numid;
    idcnt = src.idcnt;
    sgr = src.sgr;
    wit = src.wit;
    equ = src.equ;
    dcldim = src.dcldim;
    dcl = src.dcl;

    return *this;
}



//--------------------------------------------------------------------
// other operations on semigroups
//--------------------------------------------------------------------


bool  SGroup::aperiodic() const
{
    if( dcldim.size() > 0 )	// already have decomposition
    {
	vector<III_Triple>::const_iterator it;
	for( it = dcldim.begin(); it != dcldim.end(); ++it )
	    if( it->third() != 1 ) 
		return 0;
	return 1;
    }
				// owise check each element
    int  n = width();
    int  m = size();
   
    int       i,x,y,cnt,loop;
    stack<int>   S;
    uvector  indeg(n);
    uvector::const_iterator  itr;
   
    for( i = 0; i < m; i++ )
    {
	itr = semigroup().begin() + i*n;
	while( !S.empty() ) S.pop();
	cnt = loop = 0;
	Fill( indeg, 0 );
      
	for( x = 0; x < n; x++ )
	{
	    indeg[ y = *(itr+x) ]++;
	    if( x==y ) loop++; 		// vertex x has a loop
	}
      
	for( x = 0; x < n; x++ )
	    if( indeg[x]==0 ) S.push(x);

	while( !S.empty() )
	{
	    x = S.top();
	    S.pop();
	    cnt++;
	    if( --indeg[ y = *(itr+x) ]==0 )
		S.push(y);
	}
      
	if( cnt + loop < n ) return  0;
    }
    return 1;
}


//--------------------------------------------------------------------
// Printing
//--------------------------------------------------------------------

void  SGroup::print_elements( ostream& os ) const 
{
   int  ww = width();
   int  ll = size();
   int  form = ww<10 ? 2 : 3;
   uvector::const_iterator  itr = semigroup().begin();

   for( int i=0; i<ll; i++ )
   {
      os << setw(4) << i+1 << ": ";
      for( int j=0; j<ww; j++ )
	os << setw(form) << *itr++ + 1;
      os << "   " << witnesses()[i] << endl;
   }
}

void  SGroup::print_witnesses( ostream& os ) const 
{
   for( size_t i = 0; i < size(); i++ )
   {
      os << setw(4) << i+1 << ": ";
      os << witnesses()[i] << endl;
   }
}

void  SGroup::print_equations( ostream& os ) const 
{
   list<Equation>::const_iterator itr;
   os << "\n# equations: " << equations().size() << endl;
   foreach_item( itr, equations() )
     os << (*itr).first() << " -> " << (*itr).second() << endl;
   os << endl;
}

void  SGroup::print_dclasses( ostream& os ) const 
{
   for( size_t i = 0; i < numDClasses(); i++ )
   {
      os << "\n------------ D-class " << i+1 << ": ";
      os << dcldim[i] << endl;
      dcl[i].print_dclass( os );
   }
}

void  DClass::print_dclass( ostream& os ) const
{
   size_t  ww = width();
   size_t  form = ww<10 ? 2 : 3;

   for( size_t k = 0; k < kSize(); k++ )
     for( size_t i = 0; i < iSize(); i++ )
     {
	os << "---- H-class " << k << ":" << i << endl; 
	for( size_t h = 0; h < hSize(); h++ )
	{
	   for( size_t j = 0; j < ww; ++j ) 
	      os << setw(form) << get( k,i,h,j ) + 1;
	   os << endl;
	}
     }
}

ostream& operator<<(ostream& out, const SGroup& sg )
{
   out << "SGroup  \n" << sg.size() << "   " << sg.width();
   out << "  " << sg.oneflag() << endl;
   out << "#idem:   " << sg.numIdempotents() << endl;
   out << "by rank: " << sg.numIdemRank() << endl;

   if( sg.numDClasses() )
   {
      out << "D-classes: " << sg.numDClasses() << endl;
      ::PrintPlain( out, sg.dclassDimensions().begin(), sg.numDClasses() );
   }

//   sg.print_elements( out );
//   sg.print_equations( out );
//   sg.print_dclasses( out );
   return out;
}

//----------------------------------------------------------------------
// simplify rewrite rules as produced during sg generation
// produces an equivalent, reduced set of rw-rules
// Note: only suffix check necessary 

static  int   simplify_rw( list<Equation>& eq )
{
    set<string,string_lenlex>  S;      // should probably be a trie 
    string   s, t;
    size_t   i;

    list<Equation>::iterator  tit, it;

    for( it = eq.begin(); it != eq.end();  )
    {
	s = it->first();
	bool  found = 0;

	for( i = 1; !found && i < s.size(); i++ )
	{
	    t = "";
	    t.insert( 0, s, s.size()-i, s.size() );
	    if( S.find(t) != S.end() )
	    {
		found = 1;
		break;
	    }
	}

	if( !found )
	{
	    S.insert(s);		// place into set, skip ahead
	    ++it;
	}
	else
	{
	    tit = it;			// remove from eq list
	    ++tit;
	    eq.erase( it );
	    it = tit;
	}

    } // end while 
    return eq.size();
}

//----------------------------------------------------------------------

DClass::~DClass() {}

void  DClass::initialize( size_t k, size_t i, size_t h, size_t n )
{
    kdim = k;
    idim = i;
    hdim = h;
    wdth = n;
    wh   = n*h;
    whi  = wh*i;
    arr.resize( whi*k ); 
}



DClass&  DClass::operator=( const DClass& src )
{
    if( &src==this ) return *this;
    
    kdim = src.kdim;
    idim = src.idim;
    hdim = src.hdim;
    wdth = src.wdth;
    wh   = src.wh; 
    whi  = src.whi; 
    arr  = src.arr;

    return *this;
}

//----------------------------------------------------------------------

